### ANBIMA DATA ###

import pandas as pd
from requests import request, exceptions
from pprint import pprint
from stpstone.handling_data.str import StrHandler
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.cals.handling_dates import DatesBR
from stpstone.settings._global_slots import YAML_ANBIMA


class AnbimaDataGen:
    '''
    REFERENCES: https://developers.anbima.com.br/api-portal/pt-br
    '''

    def __init__(
            self,
            str_client_id: str, 
            str_client_secret: str, 
            str_env: str='dev',
            bl_debug: bool=False,
            int_chunk: int=1000,
            str_host_prd='https://api.anbima.com.br/',
            str_host_dev='https://api-sandbox.anbima.com.br/'
        ):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        self.str_client_id = str_client_id
        self.str_client_secret = str_client_secret
        self.bl_debug = bl_debug
        self.int_chunk = int_chunk
        self.str_host_prd = str_host_prd
        self.str_host = locals()[f'str_host_{str_env.lower()}']
        self.str_token = self.access_token['access_token']

    @property
    def access_token(self, str_app='oauth/access-token', str_method='POST'):
        '''
        DOCSTRING: AUTHORIZATION TOKEN IN ORDER TO GRANT ACCESS TO DESIRED APP
        INPUTS: STR_APP, STR_METHOD
        OUTPUTS: JSON
        '''
        # url
        str_url = self.str_host_prd + str_app
        # encoding credentials to base64
        base64_credentials = StrHandler().base64_encode(
            self.str_client_id, 
            self.str_client_secret
        )
        # request
        dict_headers = {
            'Content-Type': 'application/json',
            'Authorization': base64_credentials
        }
        dict_payload = {
            'grant_type': 'client_credentials'
        }
        resp_req = request(
            method=str_method, 
            url=str_url, 
            headers=dict_headers, 
            data=JsonFiles().dict_to_json(dict_payload)
        )
        if self.bl_debug == True:
            print('TOKEN: {}'.format(resp_req.json()['access_token']))
        # checking wheter response code is 2xx, otherwise raise error
        resp_req.raise_for_status()
        # returning token
        return resp_req.json()

    def generic_request(self, str_app, str_method):
        '''
        DOCSTRING: GENERIC REQUEST TO ANBIMA DATA
        INPUTS: STR_APP, STR_METHOD, DICT_PAYLOAD
        OUTPUTS: JSON
        '''
        # url
        str_url = self.str_host + str_app
        if self.bl_debug == True:
            print(f'URL: {str_url}')
        # request
        dict_headers = {
            'accept': 'application/json',
            'client_id': self.str_client_id,
            'access_token': self.str_token
        }
        resp_req = request(
            method=str_method, 
            url=str_url, 
            headers=dict_headers
        )
        # checking wheter response code is 2xx, otherwise raise error
        resp_req.raise_for_status()
        # returning token
        return resp_req.json()


class AnbimaDataFunds(AnbimaDataGen):
    '''
    REFERENCES: https://developers.anbima.com.br/pt/swagger-de-fundos-v2-rcvm-175/#/Notas%20explicativas/buscarNotasExplicativas
    '''

    def funds_raw(
        self,
        int_pg=None,
        str_app='feed/fundos/v2/fundos?size={}&page={}',
        str_method='GET'
    ):
        '''
        DOCSTRING: RETRIEVE AVAILABLE CLOSED AND OPENED-END FUNDS
        INPTUS: -
        OUTPUTS: JSON
        '''
        return self.generic_request(
            str_app.format(
                self.int_chunk,
                int_pg,
            ), 
            str_method
        )
    
    def funds_trt(self, i_pg=0, i_fnd=0):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # looping within all available fund pages
        while True:
            #   requesting for current fund json, in case status code is different
            #       from 2xx, break the loop and return the dataframe
            try:
                json_funds = self.funds_raw(i_pg)
            except exceptions.HTTPError:
                break
            # looping within content dictionaries
            for dict_cnt in json_funds[YAML_ANBIMA['anbima_data_api']['key_content']]:
                #   setting variables
                dict_aux = dict()
                i_fnd += 1
                #   looping within keys and values from content
                for YAML_ANBIMA['anbima_data_api']['key_cnt'], data_cnt in dict_cnt.items():
                    if isinstance(data_cnt, str):
                        dict_aux[YAML_ANBIMA['anbima_data_api']['key_cnt']] = data_cnt.strip()
                    elif data_cnt is None:
                        dict_aux[YAML_ANBIMA['anbima_data_api']['key_cnt']] = data_cnt
                    elif isinstance(data_cnt, list):
                        #   looping within classes
                        for i_cls, dict_cls in enumerate(data_cnt):
                            #   looping within classes and appending to serialized list
                            for YAML_ANBIMA['anbima_data_api']['key_cls'], data_cls in dict_cls.items():
                                if \
                                    (YAML_ANBIMA['anbima_data_api']['key_cls'] != YAML_ANBIMA['anbima_data_api']['key_name_sbclss'])\
                                    and (data_cls is not None):
                                    dict_aux[YAML_ANBIMA['anbima_data_api']['key_cls']] = data_cls.strip()
                                elif \
                                    (YAML_ANBIMA['anbima_data_api']['key_cls'] != YAML_ANBIMA['anbima_data_api']['key_name_sbclss']) \
                                    and (data_cls is None):
                                    dict_aux[YAML_ANBIMA['anbima_data_api']['key_cls']] = data_cls
                                elif \
                                    (YAML_ANBIMA['anbima_data_api']['key_cls'] == YAML_ANBIMA['anbima_data_api']['key_name_sbclss']) \
                                    and (isinstance(data_cls, list)):
                                    #   looping within subclasses and copy auxiliary dicitionary in each iteration, 
                                    #       in order to renew the subclass info imported
                                    for dict_sbcls in data_cls:
                                        dict_xpt = dict_aux.copy()
                                        for YAML_ANBIMA['anbima_data_api']['key_sbcls'], data_sbcls in dict_sbcls.items():
                                            dict_xpt = HandlingDicts().merge_n_dicts(
                                                dict_xpt,
                                                {
                                                    '{}_{}'.format(
                                                        YAML_ANBIMA['anbima_data_api']['key_name_sbcls'],
                                                        YAML_ANBIMA['anbima_data_api']['key_sbcls']
                                                    ): data_sbcls
                                                },
                                                {
                                                    YAML_ANBIMA['anbima_data_api']['col_num_fnd']: i_fnd + 1,
                                                    YAML_ANBIMA['anbima_data_api']['col_num_class']: i_cls + 1,
                                                    YAML_ANBIMA['anbima_data_api']['col_num_pg']: i_pg
                                                }
                                            )
                                        list_ser.append(dict_xpt)
                                elif \
                                    (YAML_ANBIMA['anbima_data_api']['key_cls'] == YAML_ANBIMA['anbima_data_api']['key_name_sbclss'])\
                                    and (data_cls is None):
                                    list_ser.append(
                                        HandlingDicts().merge_n_dicts(
                                            dict_aux,
                                            {YAML_ANBIMA['anbima_data_api']['key_cls']: data_cls},
                                            {
                                                YAML_ANBIMA['anbima_data_api']['col_num_fnd']: i_fnd + 1,
                                                YAML_ANBIMA['anbima_data_api']['col_num_class']: i_cls + 1,
                                                YAML_ANBIMA['anbima_data_api']['col_num_pg']: i_pg
                                            }
                                        )
                                    )
                                else:
                                    raise Exception('Error of content within class, please revise ' \
                                        + 'pg: {} / key: {} / data: {}'.format(
                                            i_pg,
                                            YAML_ANBIMA['anbima_data_api']['key_cls'],
                                            data_cls
                                        ))
                    else:
                        raise Exception(f'Error of content data type, please revise the data: {data_cnt}')
            #   adding iterator
            i_pg += 1
        # appending serialized list to pandas dataframe
        df_funds = pd.DataFrame(list_ser)
        # changing columns types
        for col_dt in [
            YAML_ANBIMA['anbima_data_api']['col_fund_closure_dt'],
            YAML_ANBIMA['anbima_data_api']['col_eff_dt'],
            YAML_ANBIMA['anbima_data_api']['col_incpt_dt'],
            YAML_ANBIMA['anbima_data_api']['col_closure_dt'],
            YAML_ANBIMA['anbima_data_api']['col_sbc_incpt_dt'],
            YAML_ANBIMA['anbima_data_api']['col_sbc_closure_dt'],
            YAML_ANBIMA['anbima_data_api']['col_sbc_eff_dt']
        ]:
            df_funds[col_dt].fillna(YAML_ANBIMA['anbima_data_api']['str_dt_fill_na'], inplace=True)
            df_funds[col_dt] = [
                DatesBR().str_date_to_datetime(d, YAML_ANBIMA['anbima_data_api']['str_dt_format']) 
                for d in df_funds[col_dt]
            ]
        for col_dt in [
            YAML_ANBIMA['anbima_data_api']['col_update_ts'],
            YAML_ANBIMA['anbima_data_api']['col_sbc_update_dt']
        ]:
            df_funds[col_dt].fillna(YAML_ANBIMA['anbima_data_api']['str_ts_fill_na'], inplace=True)
            df_funds[col_dt] = [
                DatesBR().timestamp_separator_string_to_datetime(d, format=YAML_ANBIMA['anbima_data_api']['str_dt_format']) 
                for d in df_funds[col_dt]
            ]
        df_funds.fillna(YAML_ANBIMA['anbima_data_api']['str_fill_na'], inplace=True)
        df_funds = df_funds.astype({
            YAML_ANBIMA['anbima_data_api']['col_fund_code']: str,
            YAML_ANBIMA['anbima_data_api']['col_type_id']: str,
            YAML_ANBIMA['anbima_data_api']['col_fund_id']: str,
            YAML_ANBIMA['anbima_data_api']['col_comp_name']: str,
            YAML_ANBIMA['anbima_data_api']['col_trade_name']: str,
            YAML_ANBIMA['anbima_data_api']['col_fund_type']: str,
            YAML_ANBIMA['anbima_data_api']['col_class_code']: str,
            YAML_ANBIMA['anbima_data_api']['col_class_id_type']: str,
            YAML_ANBIMA['anbima_data_api']['col_class_id']: str,
            YAML_ANBIMA['anbima_data_api']['col_comp_class']: str,
            YAML_ANBIMA['anbima_data_api']['col_trd_class']: str,
            YAML_ANBIMA['anbima_data_api']['col_n1_ctg']: str,
            YAML_ANBIMA['anbima_data_api']['col_subclasses']: str,
        })
        # removing duplicates
        df_funds.drop_duplicates(inplace=True)
        # returning dataframe
        return df_funds

    def fund_raw(
        self, 
        str_code_fnd: str,
        str_app='feed/fundos/v2/fundos/{}/historico',
        str_method='GET'
    ):
        '''
        DOCSTRING: RETRIEVE INFORMATION REGARDING A SPECIFIC FUND
        INPTUS: CODIGO_FUNDO FROM ANBIMA DATA INFOS REGARDING THE CHOSEN FUND
        OUTPUTS: JSON
        '''
        return self.generic_request(str_app.format(str_code_fnd), str_method)
    
    def fund_trt(
        self,
        list_code_fnds: list
    ):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        dict_dfs = dict()
        # looping within the funds codes
        for str_code_fnd in list_code_fnds:
            #   setting variables
            dict_aux = dict()
            list_ser = list()
            dict_dfs[str_code_fnd] = list()
            #   returning fund info
            json_fnd_info = self.fund_raw(str_code_fnd)
            #   looping within json content
            for key_cnt, data_cnt in json_fnd_info.items():
                #   checking data type, when is a list create the dictionary for the 
                #       serialized list to be appended into a pandas df
                if (isinstance(data_cnt, str) or data_cnt is None):
                    dict_aux[key_cnt] = data_cnt
                elif isinstance(data_cnt, list):
                    dict_xpt = dict_aux.copy()
                    for dict_data in data_cnt:
                        for key_data, data_data in dict_data.items():
                            #   checking wheter the data instance is string, or list
                            if (isinstance(data_data, str) or data_data is None):
                                dict_xpt['{}_{}'.format(
                                    key_cnt,
                                    key_data
                                )] = data_data
                            elif isinstance(data_data, list):
                                for dict_hist in data_data:
                                    dict_xpt_2 = dict_xpt.copy()
                                    for key_hist, data_hist in dict_hist.items():
                                        dict_xpt_2['{}_{}_{}'.format(
                                            key_cnt,
                                            key_data,
                                            key_hist
                                        )] = data_hist
                                    list_ser.append(dict_xpt_2)
                        #   regarding classes first-order dictionary has a key called historico_classe, which is 
                        #       a list (in 2024-11-08), it is treated separately in the code, in order to create the 
                        #       serialized list
                        if key_data != 'classes':
                            list_ser.append(dict_xpt)
                    df_ = pd.DataFrame(list_ser)
                    #   changing data types within columns
                    for col_ in list(df_.columns):
                        #   date
                        if \
                            (StrHandler().match_string_like(col_,'*data_*') == True) \
                            and (len(col_) == 10):
                            df_[col_].fillna(YAML_ANBIMA['anbima_data_api']['str_dt_fill_na'], inplace=True)
                            df_[col_] = [
                                DatesBR().str_date_to_datetime(d, YAML_ANBIMA['anbima_data_api']['str_dt_format']) 
                                for d in df_[col_]
                            ]
                        #   timestamp
                        elif \
                            (StrHandler().match_string_like(col_,'*data_*') == True) \
                            and (StrHandler().match_string_like(col_,'*T*') == True) \
                            and (len(col_) > 10):
                            df_[col_].fillna(YAML_ANBIMA['anbima_data_api']['str_ts_fill_na'], inplace=True)
                            df_[col_] = [
                                DatesBR().timestamp_separator_string_to_datetime(
                                    d, format=YAML_ANBIMA['anbima_data_api']['str_dt_format']
                                ) 
                                for d in df_[col_]
                            ]
                        #   float
                        elif \
                            (StrHandler().match_string_like(col_,'*percentual_*') == True):
                            df_[col_].fillna(YAML_ANBIMA['anbima_data_api']['str_float_fill_na'], inplace=True)
                            df_[col_] = [float(x) for x in df_[col_]]
                        #   string
                        else:
                            df_[col_].fillna(YAML_ANBIMA['anbima_data_api']['str_fill_na'], inplace=True)
                            df_[col_] = [str(x).strip() for x in df_[col_]]
                    #   appending to list of dataframe
                    dict_dfs[str_code_fnd].append(df_)
                else:
                    raise Exception('Type of data within the content of the fund {} not found, please check the code. '.format(
                        str_code_fnd
                    ) + 'TYPE: {}'.format(type(data_cnt)))
        # returning list of dataframes with fund infos
        return dict_dfs

    def fund_hist(
            self, 
            str_code_class: str,
            str_app='feed/fundos/v2/fundos/{}/historico',
            str_method='GET'
        ):
        '''
        DOCSTRING: HISTORICAL CHANGES OF REGISTRATION DATA FOR THE FUND
        INPTUS: CODIGO_FUNDO FROM ANBIMA DATA INFOS REGARDING THE CHOSEN FUND
        OUTPUTS: JSON
        '''
        return self.generic_request(str_app.format(str_code_class), str_method)
    
    def segment_investor(
            self, 
            str_code_class: str,
            str_app='feed/fundos/v2/fundos/segmento-investidor/{}/patrimonio-liquido',
            str_method='GET'
        ):
        '''
        DOCSTRING: INVESTOR SEGMENT INFORMATION, DISPLAYING THE DISTRIBUTION OF AUM 
            BY PERCENTAGE VALUES
        INPTUS: CODIGO_FUNDO FROM ANBIMA DATA INFOS REGARDING THE CHOSEN FUND
        OUTPUTS: JSON
        '''
        return self.generic_request(str_app.format(str_code_class), str_method)
    
    def time_series_fund(
            self,
            str_date_inf: str,
            str_date_sup: str,
            str_code_class: str,
            str_app='feed/fundos/v2/fundos/{}/serie-historica',
            str_method='GET'
        ):
        '''
        DOCSTRING: TIME SERIES FOR A GIVEN FUND
        INPTUS: CODIGO_FUNDO FROM ANBIMA DATA INFOS REGARDING THE CHOSEN FUND
        OUTPUTS: JSON
        '''
        dict_payload = {
            'size': self.int_chunk,
            'data-inicio': str_date_inf,
            'data-fim': str_date_sup
        }
        return self.generic_request(
            str_app.format(str_code_class, self.int_chunk), 
            str_method=str_method, 
            dict_payload=dict_payload
        )
    
    def funds_financials_dt(
            self,
            str_date_update: str,
            str_app='feed/fundos/v2/fundos/serie-historica/lote',
            str_method='GET'
        ):
        '''
        DOCSTRING: TIME SERIES OF FUNDS FOR A GIVEN DATE OF UPDATE
        INPTUS: CODIGO_FUNDO FROM ANBIMA DATA INFOS REGARDING THE CHOSEN FUND
        OUTPUTS: JSON
        '''
        dict_payload = {
            'data-atualizacao': str_date_update,
            'size': self.int_chunk
        }
        return self.generic_request(
            str_app, 
            str_method=str_method, 
            dict_payload=dict_payload
        )

    def funds_registration_data_dt(
            self,
            str_date_update: str,
            str_app='feed/fundos/v2/fundos/dados-cadastrais/lote',
            str_method='GET'
        ):
        '''
        DOCSTRING: FUNDS REGISTRATION FOR A GIVEN DATE
        INPTUS: DATE OF UPDATE
        OUTPUTS: JSON
        '''
        dict_payload = {
            'data-atualizacao': str_date_update,
            'size': self.int_chunk
        }
        return self.generic_request(
            str_app, 
            str_method=str_method, 
            dict_payload=dict_payload
        )
    
    @property
    def institutions(
            self,
            str_app='feed/fundos/v2/fundos/instituicoes',
            str_method='GET'
        ):
        '''
        DOCSTRING: INSTITUTIONS MANAGING CLOSED/OPENED-END FUNDS
        INPTUS: -
        OUTPUTS: JSON
        '''
        dict_payload = {
            'size': self.int_chunk
        }
        return self.generic_request(
            str_app, 
            str_method=str_method, 
            dict_payload=dict_payload
        )
    
    def institution(
            self,
            str_ein: str,
            str_app='feed/fundos/v2/fundos/instituicoes/{}',
            str_method='GET'
        ):
        '''
        DOCSTRING: INSTITUTIONS MANAGING CLOSED/OPENED-END FUNDS
        INPTUS: EIN (EMPLOYER IDENTIFICATION NUMBER)
        OUTPUTS: JSON
        '''
        dict_payload = {
            'size': self.int_chunk
        }
        return self.generic_request(
            str_app.format(str_ein), 
            str_method=str_method, 
            dict_payload=dict_payload
        )
    
    def explanatory_notes_fund(
            self, 
            str_code_class: str,
            str_app='feed/fundos/v2/fundos/{}/notas-explicativas',
            str_method='GET'
        ):
        '''
        DOCSTRING: EXPLANATORY NOTES FOR A GIVEN FUND
        INPTUS: CODIGO_FUNDO FROM ANBIMA DATA INFOS REGARDING THE CHOSEN FUND
        OUTPUTS: JSON
        '''
        return self.generic_request(str_app.format(str_code_class), str_method)


### CONNECTING TO MYSQL DATABASE ###

import pymysql
import json
import pandas as pd
from sqlalchemy import create_engine


class MySQLDatabase:

    def read_sql(self, host, port, database, user, password, query, timeout=7200):
        '''
        DOCSTRING:
        INPUTS: Database connection parameters and SQL query
        OUTPUTS: pandas DataFrame containing the query results
        '''
        # creating connection object
        conn = create_engine(f'mysql+pymysql://{user}:{password}@{host}:{port}/{database}',
                             connect_args={'connect_timeout': timeout})
        # return SQL result as DataFrame
        return pd.read_sql(query, con=conn)

    def engine(self, host, port, database, user, password, query, bl_insert_db=False):
        '''
        DOCSTRING:
        RUN SQL QUERIES FOR MYSQL DATABASE
        INPUTS: Database connection parameters and SQL query
        OUTPUTS: Query result (if SELECT) or success message (if INSERT/UPDATE/DELETE)
        '''
        try:
            #   making MySQL connection
            conn = pymysql.connect(
                host=host, 
                port=port, 
                user=user, 
                password=password, 
                database=database
            )
            cur = conn.cursor()
            cur.execute(query)
            #   if it's not an insert query, fetch the results
            if not bl_insert_db:
                list_rows = cur.fetchall()
            else:
                conn.commit()
                list_rows = True
        except pymysql.MySQLError as e:
            return f"Error while executing query: {e}"
        finally:
            #   close database connection
            if conn:
                cur.close()
                conn.close()
        return list_rows

    def insert_data(self, host, port, database, user, password, table_name, from_json=True,
                    json_data_path=None, json_mem=None, timeout_secs=3):
        '''
        DOCSTRING: INSERT DATA INTO MYSQL (FROM JSON FILE OR MEMORY)
        INPUTS: Database connection parameters, table name, JSON data (from file or memory)
        OUTPUTS: Insertion status message
        '''
        if from_json:
            #   defining record list from json
            if json_data_path:
                with open(json_data_path, 'r') as file:
                    record_list = json.load(file)
            elif json_mem:
                record_list = json_mem
            else:
                return 'Error: Neither JSON file path nor in-memory data provided.'
            #   create a SQL insert string
            if isinstance(record_list, list):
                # get the column names from the first record
                columns = list(record_list[0].keys())
                sql_string = f"INSERT INTO {table_name} ({', '.join(columns)}) VALUES "
                # insert data into database
                values_list = []
                for record_dict in record_list:
                    values = [
                        f"'{str(val).replace('\'', '\'\'')}'" if isinstance(val, str) 
                        else str(val) 
                        for val in record_dict.values()
                    ]
                    values_list.append(f"({', '.join(values)})")
                
                sql_string += ', '.join(values_list) + ';'
            # perform insertion
            try:
                conn = pymysql.connect(
                    host=host, 
                    port=port, 
                    user=user, 
                    password=password, 
                    database=database
                )
                cur = conn.cursor()
                cur.execute(sql_string)
                conn.commit()
                cur.close()
                conn.close()
                return 'OK'
            except pymysql.MySQLError as e:
                return f"Error while inserting data: {e}"
        else:
            return 'Error: Data insertion was not initiated through JSON.'



### HANDLING CLASSIFICATION ISSUES ###

# pypi.org libs
import matplotlib.pyplot as plt
from sklearn.datasets import fetch_openml
from sklearn.linear_model import SGDClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import confusion_matrix, precision_score, recall_score, f1_score, \
    roc_auc_score, adjusted_rand_score, silhouette_score, accuracy_score
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import KMeans
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.decomposition import PCA
from sklearn.naive_bayes import GaussianNB
# local libs
from stpstone.handling_data.lists import HandlingLists
from stpstone.handling_data.str import StrHandler


class InputsClassification:

    def fetch_sklearn_database(self, database_name='mnist_784', version=1, bl_asframe=False):
        '''
        DOCSTRING: DATASET FROM SCIKIT-LEARN, WHICH NORMALLY HAVE A SIMILAR DICTIONARY 
            STRUCTURE: DESCR (DESCRIPTION OF THE DATASET), DATA (ARRAY WITH ONE ROW PER 
            INSTANCE AND ONE COLUMN PER FEATURE), AND TARGET (AN ARRAY WITH LABELS)
        INPUTS: DATABASE NAME (MNIST_784 AS DEFAULT) AND VERSION
        OUTPUTS: DICTIONARY WITH DATA, TARGET, FEATURE NAMES, DESCR, DETAILS, CATEGORIES AND
            URL
        '''
        return fetch_openml(database_name, version=version, as_frame=bl_asframe)

    def show_image_from_dataset(self, array_instance, cmap='binary', shape=(28, 28),
                                bl_axis='off', complete_saving_path=None):
        '''
        DOCSTRING: SHOW IMAGE FROM DATASET
        INPUTS: VECTOR INSTANCE, CMAP (DEFAULT), SHAPE (DEFAULT)
        OUTPUTS: -
        '''
        # reshape image
        array_instance = array_instance.reshape(shape)
        # creating image
        plt.imshow(array_instance, cmap=cmap)
        plt.axis(bl_axis)
        # saving the plot, if is user's will
        if complete_saving_path != None:
            plt.savefig(complete_saving_path)
        # show result
        plt.show()


class Classification:

    def one_hot_vectorizer(self, list_corpus):
        '''
        DOCSTRING: ONE HOT VECTORIZER, AIMING TO ENCODE A LIST OF STRINGS AND CONVERT TO AN ARRAY
        INPUTS: CORPUS (LIST TYPE)
        OUTPUTS: DICTIONARY (LABLES AND ARRAY OF ONE HOT ENCODER)
        '''
        # list of words to vectorizer - alphabetical order
        list_labels = HandlingLists().extend_lists(
            *[x.split() for x in list_corpus])
        list_labels = [StrHandler().remove_end_period_marks(x).lower()
                       for x in HandlingLists().remove_duplicates(list_labels)]
        list_labels.sort()
        # one hot vectorizer
        one_hot_vectorizer = CountVectorizer(binary=True)
        array_one_hot = one_hot_vectorizer.fit_transform(list_corpus).toarray()
        # return dictionary with labels and one hot encoder
        return {
            'labels': list_labels,
            'one_hot_encoder': array_one_hot
        }

    def sgd_classifier(self, array_x, array_y, int_random_state_seed=5):
        '''
        DOCSTRING: TRAINING A BINARY CLASSIFIER WITH STOCHASTIC GRADIENT DESCENT (SGD), 
            TO EAVLUATE WHTEHER OR NOT A NEW ELEMENT BELONGS TO THE SUBSET, WHICH 
            HAS THE ADVANTAGE OF BEING CAPABLE OF HANDLING VERY LARGE DATASETS EFFICIENTLY
        INPUTS: ARRAY DATA, ARRAY TARGET, LIST OF DATA TO PREDICT AND RANDOM STATE SEED
            (5 AS DEFAULT)
        OUTPUTS: DICTIONARY WITH KEYS model AND PREDICTIONS (DATALABEL AND WHETHER IT 
            BELONGS TO THE ARRAY DATA OR NOT, CONSIDERING A CLASSIFICATION METHOD)
        '''
        # estimator
        model = SGDClassifier(random_state=int_random_state_seed)
        # fitting model
        model.fit(array_x, array_y)
        # array predictions
        array_y_hat = model.predict(array_x)
        # returning model fitted and predictions
        return {
            'model_fitted': model,
            'predictions': array_y_hat, 
            'adjusted_rand_score': adjusted_rand_score(array_y, array_y_hat),
            'score': model.score(array_x, array_y),
            'accuracy_score': accuracy_score(array_y, array_y_hat),
            'confusion_matrix': confusion_matrix(array_y, array_y_hat),
            'classes': model.classes_,
        }

    def svm(
            self, array_x, array_y, kernel='rbf', float_regularization_parameter=1,
            multiclass_classification_strategy='best',
            gamma='auto', int_random_state_seed=42):
        '''
        DOCSTRING: SUPPORT VECTOR MACHINE CLASSIFICATION TO PREDICT, USING ONE VERSUS ONE 
            STRATEGY, TRAINING BINARY CLASSIFIERS, GETTING THEIR DECISION SCORES FOR THE DATA, 
            AND CHOOSING THE MODEL WHICH HAS AN OPTIMIZED RESPONSE, MATCHING TARGET CLASSIFICATION 
            IN MOST CASES
        INPUTS: ARRAY X, ARRAY Y, KERNEL (LINEAR OR RBF ARE THE MOST COMMON), REGUALARIZATION 
            PARAMETER (1 AS DEFAULT), MULTICLASS CLASSIFICATION STRATEGY, GAMMA AND RANDOM STATE SEED
        OUTPUTS: DICT
        '''
        # classifier to the supporting vector machine classification, regarding multiclass
        #   classification strategy (best, ovr or ovo)
        if multiclass_classification_strategy == 'best':
            model = SVC(C=float_regularization_parameter, gamma=gamma,
                          random_state=int_random_state_seed, kernel=kernel)
        elif multiclass_classification_strategy == 'ovr':
            model = OneVsRestClassifier(
                SVC(C=float_regularization_parameter, gamma=gamma,
                    random_state=int_random_state_seed, kernel=kernel))
        elif multiclass_classification_strategy == 'ovo':
            model = OneVsOneClassifier(
                SVC(C=float_regularization_parameter, gamma=gamma,
                    random_state=int_random_state_seed, kernel=kernel))
        else:
            raise Exception(
                'Multiclass classification strategy ought be wheter best, ovr'
                + '(one-versus-the-rest) or ovo (one-versus-one), nevertheless it was declared'
                + ' {}, which is invalid, please revisit the parameter '.format(
                    multiclass_classification_strategy)
                + 'multiclass_classification_strategy')
        # fitting model
        model.fit(array_x, array_y)
        # array predictions
        array_y_hat = model.predict(array_x)
        # returning model fitted and predictions
        return {
            'model_fitted': model,
            'predictions': array_y_hat, 
            'adjusted_rand_score': adjusted_rand_score(array_y, array_y_hat),
            'score': model.score(array_x, array_y),
            'accuracy_score': accuracy_score(array_y, array_y_hat),
            'confusion_matrix': confusion_matrix(array_y, array_y_hat),
            'classes': model.classes_,
        }

    def decision_tree(self, array_x, array_y, impurity_crit='gini', 
                      float_max_depth=None, int_random_state_seed=42):
        '''
        REFERENCES: https://www.datacamp.com/tutorial/decision-tree-classification-python
        DOCSTRING: DECISION TREE CLASSIFIER
        INPUTS: ARRAY DATA, ARRAY TARGETS AND ARRAY DATA TO BE PREDICTED
        OUTPUTS: 
        '''
        # classifier
        model = DecisionTreeClassifier(criterion=impurity_crit, max_depth=float_max_depth, 
                                       random_state=int_random_state_seed)
        # fitting model
        model.fit(array_x, array_y)
        # array predictions
        array_y_hat = model.predict(array_x)
        # returning model fitted and predictions
        return {
            'model_fitted': model,
            'predictions': array_y_hat, 
            'adjusted_rand_score': adjusted_rand_score(array_y, array_y_hat),
            'score': model.score(array_x, array_y),
            'accuracy_score': accuracy_score(array_y, array_y_hat),
            'confusion_matrix': confusion_matrix(array_y, array_y_hat),
            'classes': model.classes_,
        }

    def random_forest(self, array_x, array_y, n_estimators=100, int_random_state_seed=42):
        '''
        REFERENCES: https://www.datacamp.com/tutorial/random-forests-classifier-python
        DOCSTRING: RANDOM FOREST CLASSIFIER, A.K.A. DECISION TREE ENSEMBLED
        INPUTS: ARRAY DATA, ARRAY TARGETS AND ARRAY DATA TO BE PREDICTED
        OUTPUTS: 
        '''
        # fitting model
        model = RandomForestClassifier(random_state=int_random_state_seed, n_estimators=n_estimators)
        # fitting model
        model.fit(array_x, array_y)
        # array predictions
        array_y_hat = model.predict(array_x)
        # returning model fitted and predictions
        return {
            'model_fitted': model,
            'predictions': array_y_hat, 
            'adjusted_rand_score': adjusted_rand_score(array_y, array_y_hat),
            'score': model.score(array_x, array_y),
            'accuracy_score': accuracy_score(array_y, array_y_hat),
            'confusion_matrix': confusion_matrix(array_y, array_y_hat),
            'classes': model.classes_,
        }

    def knn_classifier(self, array_x, array_y, int_n_neighbors=5):
        '''
        DOCSTRING: K NEIGHBORS CLASSIFIER
        INPUTS: ARRAY DATA, ARRAY TARGETS AND ARRAY DATA TO BE PREDICTED
        OUTPUTS: 
        '''
        # classifier
        model = KNeighborsClassifier(n_neighbors=int_n_neighbors)
        # fitting model
        model.fit(array_x, array_y)
        # array predictions
        array_y_hat = model.predict(array_x)
        # returning model fitted and predictions
        return {
            'model_fitted': model,
            'predictions': array_y_hat, 
            'adjusted_rand_score': adjusted_rand_score(array_y, array_y_hat),
            'score': model.score(array_x, array_y),
            'accuracy_score': accuracy_score(array_y, array_y_hat),
            'confusion_matrix': confusion_matrix(array_y, array_y_hat),
            'classes': model.classes_,
        }

    def k_means(self, n_clusters, array_y, array_x, int_random_state_seed=0):
        '''
        REFERENCES: https://www.hashtagtreinamentos.com/k-means-para-clusterizar-ciencia-dados?gad_source=1&gclid=Cj0KCQjwlZixBhCoARIsAIC745Bm8VTK5AMNUKTlV3TpYm6RB6ag2IGUIMEvNNYTmKmAfqN7O5vA6mwaAi6FEALw_wcB
        DOCSTRING: K-MEANS CLUSTERING FOR LABELED DATA
        INPUTS:
        OUTPUTS: DICT (LABELS - INDENTIFICATIONS OF RESPECTIVE CLUSTER, AND ADJUSTED RAND, WHICH 
            INDICATES THE SCORE OF CLUSTERIZATION N_CLUSTERS-WISE)
        '''
        # classifier
        model = KMeans(n_clusters=n_clusters, random_state=int_random_state_seed)
        # fitting model
        model.fit(array_x, array_y)
        # array predictions
        array_y_hat = model.predict(array_x)
        # returning model fitted and predictions
        return {
            'model_fitted': model,
            'predictions': array_y_hat, 
            'adjusted_rand_score': adjusted_rand_score(array_y, array_y_hat),
            'score': model.score(array_x, array_y),
            'accuracy_score': accuracy_score(array_y, array_y_hat),
            'confusion_matrix': confusion_matrix(array_y, array_y_hat),
            'classes': model.classes_,
        }

    def naive_bayes(self, array_x, array_y):
        '''
        DOCSTRING: NAIVE BAYES CLASSIFIER
        INPUTS: ARRAY DATA, ARRAY TARGETS AND ARRAY DATA TO BE PREDICTED
        OUTPUTS: DICT
        '''
        # classifier
        model = GaussianNB()
        # fitting model
        model.fit(array_x, array_y)
        # array predictions
        array_y_hat = model.predict(array_x)
        # returning model fitted and predictions
        return {
            'model_fitted': model,
            'predictions': array_y_hat, 
            'adjusted_rand_score': adjusted_rand_score(array_y, array_y_hat),
            'score': model.score(array_x, array_y),
            'accuracy_score': accuracy_score(array_y, array_y_hat),
            'confusion_matrix': confusion_matrix(array_y, array_y_hat),
            'classes': model.classes_,
        }

class ImageProcessing:

    def img_dims(self, name_path):
        '''
        REFERENCES: https://leandrocruvinel.medium.com/pca-na-mão-e-no-python-d559e9c8f053
        DOCSTRING: IMAGE DIMENSIONS
        INPUTS:
        OUTPUTS:
        '''
        # image dimensions (height x width x channels)
        return plt.imread(name_path).shape

    def pca_with_var_exp(self, name_path, var_exp):
        '''
        REFERENCES: https://leandrocruvinel.medium.com/pca-na-mão-e-no-python-d559e9c8f053
        DOCSTRING: PRINCIPAL COMPONENTS ANALYSIS WITH EXPLAINED VARIANCE RESULT FOR IMAGES
        INPUTS:
        OUTPUTS:
        '''
        # defining the model
        model = PCA(var_exp)
        # fit transform
        model = model.fit_transform(self.img_dims(name_path))
        # return lowered dimensioned matrice
        return model.inverse_transform(model)

    def plot_subplot(float_exp_var_ratio, *array_x):
        '''
        REFERENCES: https://leandrocruvinel.medium.com/pca-na-mão-e-no-python-d559e9c8f053
        DOCSTRING: PLOT THE ORIGINAL IMAGES AND THE IV VARIANCE REDUCTIONS
        INPUTS:
        OUTPUTS:
        '''
        # plot
        plt.subplot(3, 2, float_exp_var_ratio)
        plt.imshow(array_x, cmap='gray')
        plt.xticks([])
        plt.yticks([])


# corpus = ['Time flies flies like an arrow.', 'Fruit flies like a banana.']
# print(BinaryMulticlassClassification().one_hot_vectorizer(corpus))
# # output
# {'labels': ['a', 'an', 'arrow', 'banana', 'flies', 'fruit', 'like', 'time'], 'one_hot_encoder': array([[1, 1, 0, 1, 0, 1, 1],
#        [0, 0, 1, 1, 1, 1, 0]], dtype=int64)}

# corpus = ['Time flies flies like an arrow.', 'Fruit flies like a banana.']
# print(BinaryMulticlassClassification().convert_categories_from_strings_to_array(corpus, True))



### HANDLING DICTIONARIES ISSUES ###

from operator import itemgetter
from itertools import groupby
from functools import cmp_to_key
from collections import defaultdict, Counter, OrderedDict
from heapq import nsmallest, nlargest
from typing import Dict, Union, List, Any, Callable, Optional


class HandlingDicts:

    def min_val_key(self, dict_active):
        '''
        DOCSTRING: MINIMUN VALUE FOR A GIVEN SET OF VALUES IN A DICTIONARY
        INPUTS: ACTIVE DICTIONARY
        OUTPUTS: KEY, VALUE
        '''
        return min(dict_active.items(), key=itemgetter(1))

    def max_val_key(self, dict_active):
        '''
        DOCSTRING: MAXIMUN VALUE FOR A GIVEN SET OF VALUES IN A DICTIONARY
        INPUTS: ACTIVE DICTIONARY
        OUTPUTS: KEY, VALUE
        '''
        return max(dict_active.items(), key=itemgetter(1))

    def merge_n_dicts(self, *dicts):
        '''
        DOCSTRING: MERGE DICTIONARIES, FOR PYTHON 3.5+
        INPUTS: DICTIONARIES
        OUTPUTS: DICTIONARY
        '''
        dict_xpt = dict()
        for dict_ in dicts:
            dict_xpt = {**dict_xpt, **dict_}
        return dict_xpt

    def cmp(self, x, y):
        '''
        Replacement for built-in function cmp that was removed in Python 3
        Compare the two objects x and y and return an integer according to
        the outcome. The return value is negative if x < y, zero if x == y
        and strictly positive if x > y.
        https://portingguide.readthedocs.io/en/latest/comparisons.html#the-cmp-function
        '''
        return (x > y) - (x < y)

    def multikeysort(self, items, columns):
        '''
        REFERENCES: https://stackoverflow.com/questions/1143671/how-to-sort-objects-by-multiple-keys-in-python,
            https://stackoverflow.com/questions/28502774/typeerror-cmp-is-an-invalid-keyword-argument-for-this-function
        DOCSTRING: SORT A LIST OF DICTIONARIES
        INPUTS: LIST OF DICTS AND LIST OF COLUMNS, IF THERE IS A NEGATIVE (-) SIGN ON KEY, 
            IT WIL BE ORDERED IN REVERSE
        OUTPUTS: LIST OF DICTIONARIES
        '''
        comparers = [
            ((itemgetter(col[1:].strip()), -1) if col.startswith('-') else (itemgetter(
                col.strip()), 1))
            for col in columns
        ]

        def comparer(left, right):
            comparer_iter = (
                self.cmp(fn(left), fn(right)) * mult
                for fn, mult in comparers
            )
            return next((result for result in comparer_iter if result), 0)
        return sorted(items, key=cmp_to_key(comparer))

    def merge_dicts(self, list_ser, list_keys_merge=None, bl_sum_values_key=True):
        '''
        DOCSTRING: MERGE DICTS FOR EVERY KEY REPETITION
        INPUTS: FOREIGNER KEY, DICTS
        OUTPUTS: DICTIONARY
        '''
        # setting default variables
        dict_export = defaultdict(list)
        list_counter_dicts = list()
        # if list of keys to merge is none, return a list of every values for the same key
        if list_keys_merge != None:
            # iterating through dictionaries of interest an merging accordingly to foreigner key
            for dict_ in list_ser:
                for key, value in dict_.items():
                    if key in list_keys_merge:
                        dict_export[key].append(value)
                    else:
                        dict_export[key] = value
            if bl_sum_values_key == True:
                return {k: (sum(v) if isinstance(v, list) else v) for k, v in dict_export.items()}
            else:
                return dict_export
        else:
            for dict_ in list_ser:
                list_counter_dicts.append(Counter(dict_))
            return dict(sum(list_counter_dicts))

    def filter_list_ser(self, list_ser, foreigner_key, k_value, str_filter_type:str='equal') \
        -> List[Dict[str, Any]]:
        '''
        DOCSTRING: FILTER LIST OF DICTIONARIES
        INPUTS: 
            - LIST OF DICTS: LIST
            - FOREINGER KEY: STRING
            - VALUE OF FOREIGNER KEY OF INTEREST: STRING/INTEGER/FLOAT
            - TYPE OF FILTER: STR
        OUTPUTS: LIST OF DICTS
        '''
        if str_filter_type == 'equal':
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] == k_value]
        elif str_filter_type == 'not_equal':
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] != k_value]
        elif str_filter_type == 'less_than':
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] < k_value]
        elif str_filter_type == 'greater_than':
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] > k_value]
        elif str_filter_type == 'less_than_or_equal_to':
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] <= k_value]
        elif str_filter_type == 'greater_than_or_equal_to':
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] >= k_value]
        elif str_filter_type == 'isin':
            return [dict_ for dict_ in list_ser if dict_[foreigner_key] in k_value]
        else:
            raise ValueError(
                'str_filter_type must be "equal", "not_equal", "less_than" or "greater_than"')

    def merge_values_foreigner_keys(self, list_ser, foreigner_key, list_keys_merge_dict):
        '''
        REFERECES: https://stackoverflow.com/questions/50167565/python-how-to-merge-dict-in-list-of-dicts-based-on-value
        DOCSTRING: MERGE DICTS ACCORDINGLY TO A FOREIGNER KEY IN THE LIST OF DICTS
        INPUTS: LIST OF DICTS, FOREIGNER KEY AND LIST OF KEYS TO MERGE IN A GIVEN SET O DICTS
        OUTPUTS: LIST OF DICTS
        '''
        # setting default variables
        list_ser_export = list()
        list_foreinger_keys = list()
        list_ser_export = list()
        # get values from foreinger key
        list_foreinger_keys = list(
            set([dict_[foreigner_key] for dict_ in list_ser]))
        # iterating through list of foreigner key and merging values of interest
        for key in list_foreinger_keys:
            # filter dicts for the given foreinger key
            list_filtered_dicts = self.filter_list_ser(
                list_ser, foreigner_key, key)
            # merge dictionaries accordingly to given keys
            list_ser_export.append(self.merge_dicts(
                list_filtered_dicts, list_keys_merge_dict))
        # return final result
        return list_ser_export

    def n_smallest(self, list_ser, key_, n):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return nsmallest(n, list_ser, key=lambda dict_: dict_[key_])

    def n_largest(self, list_ser, key_, n):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return nlargest(n, list_ser, key=lambda dict_: dict_[key_])

    def order_dict(self, dict_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return OrderedDict(sorted(dict_.items()))

    def group_by_dicts(self, list_ser):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # sort by the desired field first
        list_ser.sort(key=itemgetter('date'))
        # return iteration in groups
        return groupby(list_ser, key=itemgetter('date'))

    def add_key_value_to_dicts(
        self, list_ser:List[Dict[str, Union[int, float, str]]], key:str,
        value:Union[Callable[..., Union[int, float, str]], Union[int, float, str]],
        list_keys_for_function:Optional[List[str]]=None, 
        kwargs_static:Optional[Dict[str, Union[int, float, str, None]]]=None
    ) -> List[Dict[str, Union[int, float, str]]]:
        '''
        DOCSTRING: ADDS A KEY AND VALUE TO EVERY DICTIONARY IN A LIST
        INPUTS:
            LIST_SER: A LIST OF DICTIONARIES TO BE UPDATED - LIST SERIALIZED
            KEY: THE KEY TO ADD TO EACH DICTIONARY
            VALUE: THE VALUE TO ASSOCIATE WITH THE KEY OR A FUNCTION TO COMPUTE THE VALUE
            KEYS_FOR_FUNCTION: KEYS TO EXTRACT FROM THE DICTIONARY FOR THE VALUE FUNCTION (IF CALLABLE)
        OUTPUTS:
            LIST OF UPDATED DICTIONARIES.
        '''
        for dict_ in list_ser:
            if isinstance(dict_, dict):
                #   check if the value is a function, otherwise assign the static value
                if callable(value):
                    args = [dict_.get(k) for k in list_keys_for_function] \
                        if list_keys_for_function is not None else []
                    if kwargs_static is not None:
                        dict_[key] = value(*args, **kwargs_static)
                    else:
                        dict_[key] = value(*args)
                else:
                    dict_[key] = value
        return list_ser

    def pair_headers_with_data(self, list_headers, list_data):
        '''
        DOCSTRING: PAIR HEADERS AND DATA AS KEYS AND VALUES IN A SERIALIZED LIST
            - FOR EXAMPLE, IF LIST_HEADERS IS ['NAME', 'AGE'] AND LIST_DATA IS 
                ['JOHN', 25, 'ALICE', 30], THE FUNCTION WILL RETURN [{'NAME': 'JOHN', 'AGE': 25}, 
                {'NAME': 'ALICE', 'AGE': 30}]
        INPUTS: LIST HEADERS, LIST DATA
        OUTPUTS: LIST
        '''
        # setting variables
        list_ser = list()
        # ensuring the list_data length is a multiple of list_headers length
        if len(list_data) % len(list_headers) != 0:
            raise ValueError(
                'The length of list_data is not a multiple of the length of list_headers.')
        # iterate over the list_data in chunks equal to the length of list_headers
        for i in range(0, len(list_data), len(list_headers)):
            # create a dictionary for each chunk
            entry = {list_headers[j]: list_data[i + j] for j in range(len(list_headers))}
            list_ser.append(entry)
        # returning list of dictionaries
        return list_ser

    def replace_variables(dict_base:Dict[str, Any], dict_replacer:Dict[str, Any]) -> Dict[str, Any]:
        '''
        DOCSTRING: REPLACE PLACEHOLDERS IN THE BASE DICTIONARY WITH VALUES FROM THE REPLACEMENT 
            DICTIONARY
        INPUTS:
            - DICT_BASE:DICT - THE DICTIONARY CONTAINING PLACEHOLDERS.
            - DICT_REPLACER:DICT - THE DICTIONARY PROVIDING REPLACEMENT VALUES.
        OUTPUTS: DICT
        '''
        for key, value in dict_base.items():
            if isinstance(value, dict):
                #   handle nested dictionaries
                dict_base[key] = dict_replacer(value, dict_replacer)
            elif isinstance(value, str) and value.startswith('{{') and value.endswith('}}'):
                #   extract variable name and replace with corresponding value
                var_name = value.strip('{{ }}').strip()
                if var_name in dict_replacer:
                    dict_base[key] = dict_replacer[var_name]
        return dict_base


### BIBLIOTECA COM FUNÇÕES PARA USAR COM O EXCEL ###
'''
### fontes:  ###
https://gist.github.com/mikepsn/27dd0d768ccede849051#file-excelapp-py-L16
http://code.activestate.com/recipes/528870-class-for-writing-content-to-excel-and-formatting-/
'''
from __future__ import unicode_literals
import os
import io
import pandas as pd
from win32com.client import Dispatch, constants
from xlwt import Workbook
from ctypes import windll
from stpstone.handling_data.folders import DirFilesManagement


STYLE_HEADING1 = "style_heading1"
STYLE_HEADING2 = "style_heading2"
STYLE_BORDER_BOTTOM = "style_border_bottom"
STYLE_GREY_CELL = "style_grey_cell"
STYLE_PALE_YELLOW_CELL = "style_pale_yellow_cell"
STYLE_ITALICS = "style_italics"

# these are the constant values in one particular version of EXCEL - if having problems,
# check your own
XL_CONST_EDGE_LEFT = 7
XL_CONST_EDGE_BOTTOM = 9
XL_CONST_CONTINUOUS = 1
XL_CONST_AUTOMATIC = -4105
XL_CONST_THIN = 2
XL_CONST_GRAY16 = 17
XL_CONST_SOLID = 1

# Debug.Print RGB(230,230,230) in Excel Immediate window
RGB_PALE_GREY = 15132390
RGB_PALE_YELLOW = 13565951  # RGB(255,255,206)


class DealingExcel:

    def save_as(self, active_wb, nome_comp_arq):
        '''
        DOCSTRING: SALVAR COMO EXCEL DE INTERESSE
        INPUTS: WORKBOOK ATIVO E NOME COMPLETO DO ARQUIVO XLS
        OUTPUTS: -
        '''
        xlapp = Dispatch('Excel.Application')
        xlapp.Visible = 0
        wb = xlapp.Workbooks.Open(active_wb)
        wb.SaveAs(nome_comp_arq)
        wb.Close(True)

    def save_as_active_wb(self, filename, filename_as):
        '''
        DOCSTRING: SAVE AS ACTIVE WORKBOOK
        INPUTS: WORKBOOK NAME AND NAME TO SAVE AS
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        '''
        xlapp = Dispatch('Excel.Application')
        xlapp.Visible = 0
        wb = xlapp.Workbooks.Open(filename)
        wb.SaveAs(filename_as)
        wb.Close(True)

        return DirFilesManagement().object_exists(filename_as)

    def open_xl(self, nome_comp_arq):
        '''
        DOCSTRING: ABRIR EXCEL DE INTERESSE
        INPUTS: NOME DO ARQUIVO DE INTERESSE
        OUTPUTS: EXCEL APP E PASTA DE TRABALHO ABERTA
        '''
        xlapp = Dispatch('Excel.Application')
        xlapp.Visible = 0
        wb = xlapp.Workbooks.Open(nome_comp_arq)

        return xlapp, wb

    def close_wb(self, active_wb, save_orn=True):
        '''
        DOCSTRING: CLOSE ACTIVE WORKBOOK
        INPUTS: ACTIVE WORKBOOK AND WHETER SAVE IT OR NOT (BOOLEAN)
        OUTPUTS: -
        '''
        active_wb.Close(save_orn)

    def delete_sheet(self, sheet_name, excel_app, active_wb):
        '''
        DOCSTRING: DELETE A SHEET FROM THE WORKBOOK
        INPUTS: SHEET NAME, XP APP AND WORBOOK
        '''
        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        sheets = active_wb.Sheets
        excel_app.DisplayAlerts = False
        sheets(sheet_name).Delete()
        excel_app.DisplayAlerts = True

    def paste_values_column(self, nome_plan, str_range, excel_app, active_wb):
        '''
        DOCSTRING: COLAR COMO VALOR COLUNA DE INTERESSE
        INPUTS: COLUNA COMO STR
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        excel_app.Range(str_range).Copy()
        excel_app.Range(str_range).PasteSpecial(Paste=-4163)
        DealingExcel().clearclipboard()

    def color_range_w_rule(self, nome_plan, str_range, matching_value, cor_pintar_intervalo,
                           excel_app, active_wb):
        '''
        DOCSTRING: COLORIR TODA A LINHA
        INPUTS: NOME DA PLAN, COLUNA COMO STR, XL APP E WORKBOOK ATIVO
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # Selecionar apenas números no intervalo
        for cel in excel_app.ActiveSheet.Range(str_range).SpecialCells(2):
            if str(cel) == matching_value:
                cel.EntireRow.Interior.Pattern = 1
                cel.EntireRow.Interior.PatternColorIndex = -4105
                cel.EntireRow.Interior.Color = cor_pintar_intervalo
                cel.EntireRow.Interior.TintAndShade = 0
                cel.EntireRow.Interior.PatternTintAndShade = 0

    def autofit_range_columns(self, nome_plan, str_range, excel_app, active_wb):
        '''
        DOCSTRING: COLAR COMO VALOR COLUNA DE INTERESSE
        INPUTS: NOME DA PLAN, COLUNA COMO STR, XL APP E WORKBOOK ATIVO
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # Autofit column in active sheet
        excel_app.ActiveSheet.Columns(str_range).Columns.AutoFit()

    def delete_entire_column(self, nome_plan, str_range, excel_app, active_wb):
        '''
        DOCSTRING: DELETAR COLUNA DE UMA PLANILHA EXCEL
        INPUTS: NOME COMPLETO DO ARQUIVO, PLAN DE INTERESSE, RANGE DE INTERESSE (COLUNA) , 
        XL APP E WORKBOOK ATIVO
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # Delete column of interest
        excel_app.ActiveSheet.Range(str_range).Columns.Delete()

    def delete_cells_w_specific_data(self, nome_plan, str_range, value_to_delete,
                                     excel_app, active_wb):
        '''
        DOCSTRING: DELETAR VALORES DE UMA COLUNA CASO COINCIDAM COM UM VALOR ESPECÍFICO
        INPUTS: RANGE DE INTERESSE (COLUNA) , NOME PLANILHA, VALOR DE INTERESSE A SER DELETADO,
            XL APP E WORKBOOK ATIVO
        '''
        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # deletar dado de interesse em range
        for cel in excel_app.ActiveSheet.Range(str_range).SpecialCells(2):
            if str(cel) == value_to_delete:
                excel_app.ActiveSheet.Cells(
                    cel.row, cel.column + 1).value = 'DELETAR'

        excel_app.Worksheets(nome_plan).Cells.Find('DELETAR').EntireColumn.SpecialCells(
            2).EntireRow.Delete

    def replacing_all_matching_str_column(self, nome_plan, str_range, value_to_replace,
                                          replace_value, excel_app, active_wb):
        '''
        DOCSTRING: SUBSTITUIR TODOS OS VALORES NUMÉRICOS EM UMA COLUNA DE UMA ARQUIVO EXCEL
        INPUTS: NOME COMPLETO DO ARQUIVO, PLAN DE INTERESSE E RANGE DE INTERESSE (COLUNA)
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # Selecionar apenas números no intervalo
        for cel in excel_app.ActiveSheet.Range(str_range).SpecialCells(2):
            if str(cel) == value_to_replace:
                cel.value = replace_value

    def create_new_column(self, nome_plan, str_range, excel_app, active_wb):
        '''
        DOCSTRING: CRIAR COLUNA EM UMA PLANILHA EXCEL
        INPUTS: NOME COMPLETO DO ARQUIVO, PLAN DE INTERESSE, RANGE DE INTERESSE (COLUNA) ANTES DA QUAL
        SERÁ CRIADA A COLUNA, XL APP E WORKBOOK ATIVO
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # Delete column of interest
        excel_app.ActiveSheet.Columns(str_range).Insert()

    def text_to_columns(self, nome_plan, str_range, excel_app, active_wb):
        '''
        DOCSTRING: CRIAR COLUNA EM UMA PLANILHA EXCEL
        INPUTS: NOME COMPLETO DO ARQUIVO, PLAN DE INTERESSE, RANGE DE INTERESSE (COLUNA) ANTES DA QUAL
        SERÁ CRIADA A COLUNA, XL APP E WORKBOOK ATIVO
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # Delete column of interest
        excel_app.ActiveSheet.Range(str_range).TextToColumns()

    def conf_title(self, nome_plan, str_range, excel_app, active_wb):
        '''
        DOCSTRING: CONFIGURE 
        INPUTS: NOME COMPLETO DO ARQUIVO, PLAN DE INTERESSE, RANGE DE INTERESSE (COLUNA) ANTES DA QUAL
        SERÁ CRIADA A COLUNA, XL APP E WORKBOOK ATIVO
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # Changing borders
        excel_app.Range(str_range).Borders(-4131).LineStyle = -4142
        excel_app.Range(str_range).Borders(-4152).LineStyle = -4142
        excel_app.Range(str_range).Borders(-4160).LineStyle = -4142
        excel_app.Range(str_range).Borders(-4107).LineStyle = -4142
        excel_app.Range(str_range).Borders(5).LineStyle = -4142
        excel_app.Range(str_range).Borders(6).LineStyle = -4142

        # Bolding text font
        excel_app.Range(str_range).Font.Bold = True

    def copy_column_paste_in_other(self, nome_plan, str_range_a, str_range_b, excel_app, active_wb):
        '''
        DOCSTRING: CRIAR COLUNA EM UMA PLANILHA EXCEL
        INPUTS: NOME COMPLETO DO ARQUIVO, PLAN DE INTERESSE, RANGE DE INTERESSE (COLUNA) ANTES DA QUAL
        SERÁ CRIADA A COLUNA, XL APP E WORKBOOK ATIVO
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # copy & paste column
        excel_app.Range(str_range_a).Copy()
        excel_app.Range(str_range_b).PasteSpecial(-4104)
        DealingExcel().clearclipboard()

    def move_column(self, nome_plan, str_range_a, str_range_b, excel_app, active_wb):
        '''
        DOCSTRING: CRIAR COLUNA EM UMA PLANILHA EXCEL
        INPUTS: NOME COMPLETO DO ARQUIVO, PLAN DE INTERESSE, RANGE DE INTERESSE (COLUNA) ANTES DA QUAL
        SERÁ CRIADA A COLUNA, XL APP E WORKBOOK ATIVO
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # copy & paste column
        excel_app.Columns(str_range_a).Cut()
        excel_app.Columns(str_range_b).Insert()
        DealingExcel().clearclipboard()

    def attr_value_cel(self, nome_plan, str_range, value_inter, excel_app, active_wb):
        '''
        DOCSTRING: CHANGE A CEL VALUE
        INPUTS: NOME PLAN INTERESSE, RANGE (CEL), NEW VALUE, XL APP E ACTIVEWORKBOOK
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # attribute value
        excel_app.Range(str_range).Value = value_inter

    def number_format(self, nome_plan, str_range, format_inter, excel_app, active_wb):
        '''
        DOCSTRING: CHANGE A CEL VALUE
        INPUTS: NOME PLAN INTERESSE, RANGE (CEL), NEW VALUE, XL APP E ACTIVEWORKBOOK
        OUTPUTS: -
        '''

        # activate workbook
        active_wb.Activate()

        # Activate sheet of interest
        excel_app.Worksheets(nome_plan).Activate()

        # attribute value
        excel_app.Range(str_range).NumberFormat = format_inter

    def clearclipboard(self):
        if windll.user32.OpenClipboard(None):
            windll.user32.EmptyClipboard()
            windll.user32.CloseClipboard()

    def restore_corrupted_xl(self, filename_orig, filename_output):
        # Opening the file using 'utf-16'/"iso-8859-1" encoding
        file1 = io.open(filename_orig, "r", encoding='utf-16')
        data = file1.readlines()

        # Creating a workbook object
        xldoc = Workbook()
        # Adding a sheet to the workbook object
        sheet = xldoc.add_sheet("Sheet1", cell_overwrite_ok=True)
        # Iterating and saving the data to sheet
        for i, row in enumerate(data):
            # Two things are done here
            # Removeing the '\n' which comes while reading the file using io.open
            # Getting the values after splitting using '\t'
            for j, val in enumerate(row.replace('\n', '').split('\t')):
                sheet.write(i, j, val)

        # Saving the file as an excel file
        xldoc.save(filename_output)

        return DirFilesManagement().object_exists(filename_output)


# excelchart
# creates a microsoft excel chart given a data range
# and whole bunch of other parameters


class ExcelChart:
    def __init__(self, excel, workbook, chartname, afterSheet):
        self.chartname = chartname
        self.excel = excel
        self.workbook = workbook
        self.chartname = chartname
        self.afterSheet = afterSheet

    def SetTitle(self, chartTitle):
        self.chartTitle = chartTitle

    def SetType(self, chartType):
        self.chartType = chartType

    def SetSource(self, chartSource):
        self.chartSource = chartSource

    def SetPlotBy(self, plotBy):
        self.plotBy = plotBy

    def SetCategoryLabels(self, numCategoryLabels):
        self.numCategoryLabels = numCategoryLabels

    def SetSeriesLabels(self, numSeriesLabels):
        self.numSeriesLabels = numSeriesLabels

    def SetCategoryTitle(self, categoryTitle):
        self.categoryTitle = categoryTitle

    def SetValueTitle(self, valueTitle):
        self.valueTitle = valueTitle

    def CreateChart(self):
        self.chart = self.workbook.Charts.Add(After=self.afterSheet)
        self.chart.ChartWizard(Gallery=constants.xlColumn,
                               CategoryLabels=1,
                               SeriesLabels=1,
                               CategoryTitle=self.categoryTitle,
                               ValueTitle=self.valueTitle,
                               PlotBy=self.plotBy,
                               Title=self.chartTitle)
        self.chart.SetSourceData(self.chartSource, self.plotBy)
        self.chart.HasAxis = (constants.xlCategory, constants.xlPrimary)
        self.chart.Axes(constants.xlCategory).HasTitle = 1
        self.chart.Axes(
            constants.xlCategory).AxisTitle.Text = self.categoryTitle
        self.chart.Axes(constants.xlValue).HasTitle = 1
        self.chart.Axes(constants.xlValue).AxisTitle.Text = self.valueTitle
        self.chart.Axes(
            constants.xlValue).AxisTitle.Orientation = constants.xlUpward
        self.chart.PlotBy = self.plotBy
        self.chart.Name = self.chartname
        self.chart.HasTitle = 1
        self.chart.ChartTitle.Text = self.chartTitle
        self.chart.HasDataTable = 0
        self.chart.ChartType = self.chartType

    def SetLegendPosition(self, legendPosition):
        self.chart.Legend.Position = legendPosition

    def PlotByColumns(self):
        self.chart.PlotBy = constants.xlColumns

    def PlotByRows(self):
        self.chart.PlotBy = constants.xlRows

    def SetCategoryAxisRange(self, minValue, maxValue):
        self.chart.Axes(constants.xlCategory).MinimumScale = minValue
        self.chart.Axes(constants.xlCategory).MaximumScale = maxValue

    def SetValueAxisRange(self, minValue, maxValue):
        self.chart.Axes(constants.xlValue).MinimumScale = minValue
        self.chart.Axes(constants.xlValue).MaximumScale = maxValue

    def ApplyDataLabels(self, dataLabelType):
        self.chart.ApplyDataLabels(dataLabelType)

    def SetBorderLineStyle(self, lineStyle):
        self.chart.PlotArea.Border.LineStyle = lineStyle

    def SetInteriorStyle(self, interiorStyle):
        self.chart.PlotArea.Interior.Pattern = interiorStyle

# excelworksheet
# creates an excel worksheet


class ExcelWorksheet:
    def __init__(self, excel, workbook, sheetname):
        self.sheetname = sheetname
        self.excel = excel
        self.workbook = workbook
        self.worksheet = self.workbook.Worksheets.Add()
        self.worksheet.Name = sheetname

    def Activate(self):
        self.worksheet.Activate()

    def SetCell(self, row, col, value):
        self.worksheet.Cells(row, col).Value = value

    def GetCell(self, row, col):
        return self.worksheet.Cells(row, col).Value

    def SetFont(self, row, col, font, size):
        self.worksheet.Cells(row, col).Font.Name = font
        self.worksheet.Cells(row, col).Font.Size = size

    def GetFont(self, row, col):
        font = self.worksheet.Cells(row, col).Font.Name
        size = self.worksheet.Cells(row, col).Font.Size
        return (font, size)

# excelWorkbook
# creates an Excel Workbook


class ExcelWorkbook:
    def __init__(self, excel, filename):
        self.filename = filename
        self.excel = excel
        self.workbook = self.excel.Workbooks.Add()
        self.worksheets = {}

    def addworksheet(self, name):
        worksheet = ExcelWorksheet(self.excel, self.workbook, name)
        self.worksheets[name] = worksheet
        return worksheet

    def addchart(self, name, afterSheet):
        chart = ExcelChart(self.excel, self.workbook, name, afterSheet)
        self.worksheets[name] = chart
        return chart

    def save(self):
        self.workbook.SaveAs(self.filename)

    def close(self):
        self.worksheets = {}
        self.workbook.Close()

    def setauthor(self, author):
        self.workbook.Author = author

# excelapp
# encapsulates an Excel Application


class ExcelApp:
    def __init__(self):
        self.excel = Dispatch("Excel.Application")
        self.workbooks = []
        self.SetDefaultSheetNum(1)

    def Show(self):
        self.excel.Visible = 1

    def Hide(self):
        self.excel.Visible = 0

    def Quit(self):
        for wkb in self.workbooks:
            wkb.Close()
        self.excel.Quit()

    def SetDefaultSheetNum(self, numSheets):
        self.excel.SheetsInNewWorkbook = numSheets

    def AddWorkbook(self, filename):
        workbook = ExcelWorkbook(self.excel, filename)
        self.workbooks.append(workbook)
        return workbook


class ExcelWriter(object):
    """
    Excel class for creating spreadsheets - esp writing data and formatting them
    Based in part on #http://snippets.dzone.com/posts/show/2036,
    and http://www.markcarter.me.uk/computing/python/excel.html
    """

    def __init__(self, file_name, default_sheet_name, make_visible=False):
        """Open spreadsheet"""
        self.excelapp = Dispatch("Excel.Application")
        if make_visible:
            self.excelapp.Visible = 1  # fun to watch!
        self.excelapp.Workbooks.Add()
        self.workbook = self.excelapp.ActiveWorkbook
        self.file_name = file_name
        self.default_sheet = self.excelapp.ActiveSheet
        self.default_sheet.Name = default_sheet_name

    def getExcelApp(self):
        """Get Excel App for use"""
        return self.excelapp

    def addSheetAfter(self, sheet_name, index_or_name):
        """
        Add new sheet to workbook after index_or_name (indexing starts at 1).
        """
        sheets = self.workbook.Sheets
        # Sheets.Add(Before, After, Count, Type) - http://www.functionx.com/vbaexcel/Lesson07.htm
        sheets.Add(None, sheets(index_or_name)).Name = sheet_name

    def deleteSheet(self, sheet_name):
        """Delete named sheet"""
        # http://www.exceltip.com/st/Delete_sheets_without_confirmation_prompts_using_VBA_in_Microsoft_Excel/483.html
        sheets = self.workbook.Sheets
        self.excelapp.DisplayAlerts = False
        sheets(sheet_name).Delete()
        self.excelapp.DisplayAlerts = True

    def getSheet(self, sheet_name):
        """
        Get sheet by name.
        """
        return self.workbook.Sheets(sheet_name)

    def activateSheet(self, sheet_name):
        """
        Activate named sheet.
        """
        sheets = self.workbook.Sheets
        # http://mail.python.org/pipermail/python-win32/2002-February/000249.html
        sheets(sheet_name).Activate()

    def add2cell(self, row, col, content, sheet=None):
        """
        Add content to cell at row,col location.
        NB only recommended for small amounts of data http://support.microsoft.com/kb/247412.
        """
        if sheet == None:
            sheet = self.default_sheet
        sheet.Cells(row, col).Value = content

    def addRow(self, row_i, data_tuple, start_col=1, sheet=None):
        """
        Add row in a single operation.  Takes a tuple per row.
        Much more efficient than cell by cell. http://support.microsoft.com/kb/247412.
        """
        if sheet == None:
            sheet = self.default_sheet
        col_n = len(data_tuple)
        last_col = start_col + col_n - 1
        insert_range = self.getRangeByCells(
            (row_i, start_col), (row_i, last_col), sheet)
        insert_range.Value = data_tuple

    def addMultipleRows(self, start_row, list_data_tuples, start_col=1, sheet=None):
        """
        Adds data multiple rows at a time, not cell by cell. Takes list of tuples
        e.g. cursor.fetchall() after running a query
        One tuple per row.
        Much more efficient than cell by cell or row by row.
        http://support.microsoft.com/kb/247412.
        Returns next available row.
        """
        if sheet == None:
            sheet = self.default_sheet
        row_n = len(list_data_tuples)
        last_row = start_row + row_n - 1
        col_n = len(list_data_tuples[0])
        last_col = start_col + col_n - 1
        insert_range = self.getRangeByCells(
            (start_row, start_col), (last_row, last_col), sheet)
        insert_range.Value = list_data_tuples
        next_available_row = last_row + 1

        return next_available_row

    def getRangeByCells(self, cell_start_row, cell_start_col, cell_end_row, cell_end_col,
                        sheet=None):
        """Get a range defined by cell start and cell end e.g. (1,1) A1 and (7,2) B7"""
        if sheet == None:
            sheet = self.default_sheet

        return sheet.Range(sheet.Cells(cell_start_row, cell_start_col),
                           sheet.Cells(cell_end_row, cell_end_col))

    def fitCols(self, col_start, col_end, sheet=None):
        """
        Fit colums to contents.
        """
        if sheet == None:
            sheet = self.default_sheet
        col_n = col_start
        while col_n <= col_end:
            self.fitCol(col_n, sheet)
            col_n = col_n + 1

    def fitCol(self, col_n, sheet=None):
        """
        Fit column to contents.
        """
        if sheet == None:
            sheet = self.default_sheet
        sheet.Range(sheet.Cells(1, col_n), sheet.Cells(
            1, col_n)).EntireColumn.AutoFit()

    def setColWidth(self, col_n, width, sheet=None):
        """
        Set column width.
        """
        if sheet == None:
            sheet = self.default_sheet
        sheet.Range(sheet.Cells(1, col_n), sheet.Cells(
            1, col_n)).ColumnWidth = width

    def formatRange(self, range, style):
        """
        Add formatting to a cell/group of cells.
        To get methods etc record a macro in EXCEL and look at it.
        To get the value of Excel Constants such as xlEdgeLeft (7) or xlThin (2)
        type e.g. Debug.Print xlEdgeLeft in the Immediate window of the VBA editor and press enter.
        http://www.ureader.com/message/33389340.aspx
        For changing the pallete of 56 colours ref: http://www.cpearson.com/excel/colors.htm    
        """
        if style == STYLE_HEADING1:
            range.Font.Bold = True
            range.Font.Name = "Arial"
            range.Font.Size = 12
        elif style == STYLE_HEADING2:
            range.Font.Bold = True
            range.Font.Name = "Arial"
            range.Font.Size = 10.5
        elif style == STYLE_BORDER_BOTTOM:
            range.Borders(XL_CONST_EDGE_BOTTOM).LineStyle = XL_CONST_CONTINUOUS
            range.Borders(XL_CONST_EDGE_BOTTOM).Weight = XL_CONST_THIN
            range.Borders(XL_CONST_EDGE_BOTTOM).ColorIndex = XL_CONST_AUTOMATIC
        elif style == STYLE_GREY_CELL:
            self.resetColorPallet(1, RGB_PALE_GREY)
            range.Interior.ColorIndex = 1
            range.Interior.Pattern = XL_CONST_SOLID
        elif style == STYLE_PALE_YELLOW_CELL:
            self.resetColorPallet(1, RGB_PALE_YELLOW)
            range.Interior.ColorIndex = 1
            range.Interior.Pattern = XL_CONST_SOLID
        elif style == STYLE_ITALICS:
            range.Font.Italic = True
        else:
            raise Exception("Style '%s' has not been defined" % style)

    def resetColorPallet(self, color_index, color):
        """
        Reset indexed color in pallet (limited to 1-56).
        Get color values by Debug.Print RGB(230,230,230) in Excel Immediate window
        """
        if color_index < 1 or color_index > 56:
            raise Exception(
                "Only indexes between 1 and 56 are available in the Excel color pallet.")
        colors_tup = self.workbook.Colors  # immutable of course
        colors_list = list(colors_tup)
        # zero-based in list but not Excel pallet
        colors_list[color_index - 1] = RGB_PALE_GREY
        new_colors_tup = tuple(colors_list)
        self.workbook.Colors = new_colors_tup

    def mergeRange(self, range):
        """Merge range"""
        range.Merge()

    def save(self):
        """Save spreadsheet as filename - wipes if existing"""
        if os.path.exists(self.file_name):
            os.remove(self.file_name)
        self.workbook.SaveAs(self.file_name)

    def close(self):
        """Close spreadsheet resources"""
        self.workbook.Saved = 0  # p.248 Using VBA 5
        self.workbook.Close(SaveChanges=0)  # to avoid prompt
        self.excelapp.Quit()
        # must make Visible=0 before del self.excelapp or EXCEL.EXE remains in memory.
        self.excelapp.Visible = 0
        del self.excelapp



# OPTION PRCING FORMULAS, FOR AMERICAN TYPE

import numpy as np


class InitialSettings:

    def set_parameters(self, *params, opt_style='call'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # check wheter is a call or put option, in case the type is neither of the former raise
        #   error
        if opt_style not in ['call', 'put']:
            raise Exception('Option ought be a call or a put')
        # initial parameters
        list_params = [param[0] if isinstance(param, list) == True else param for
                       param in params]
        # return parameters
        return [float(param) for param in list_params if isinstance(param, str) == False]


class PricingModels(InitialSettings):

    def binomial(self, s, k, r, t, n, u, d, opt_style, h_upper=None, h_lower=None):
        '''
        REFERENCES: 
            https://www.youtube.com/watch?v=a3906k9C0fM, 
            https://www.youtube.com/watch?v=WxrRi9lNnqY,
            https://www.youtube.com/watch?v=K2Iy8bCmXjk
        DOCSTRING: BINOMIAL ASSET PRICING MODEL - UPPER/LOWER BARRIERS FEATURES IMPLEMENTED
        INPUTS: S0 (INITIAL STOCK PRICING MODEL), K (STRIKE PRICE), T (TIME TO MATURITY IN YEARS), 
            R (ANNUAL RISK-FREE RATE), N (NODES), U (UP-FACTOR IN BINOMIAL MODELS), D (DOWN-FACTOR 
            - TO ENSURE RECOMBINING TREE USE 1/U), OPTION STYLE (CALL/PUT)
        OUTPUTS: FLOAT FROM ARRAY_CP[0] (CALL-PUT PRICING FOR EACH NODE)
        '''
        # initial parameters
        s, k, r, t, n, u, d = self.set_parameters(
            s, k, r, t, n, u, d, opt_style)
        # precompute values
        dt = t / n
        q = (np.exp(r * dt) - d) / (u - d)
        disc = np.exp(-r * dt)
        # initialise stock prices at maturity
        array_s_nodes = s * d**(np.arange(n, -1, -1)) * \
            u**(np.arange(0, n + 1, 1))
        # initialise option values at maturity - if intrinsic value is negative, consider zero
        if opt_style == 'call':
            array_cp = np.maximum(array_s_nodes - k, np.zeros(int(n) + 1))
        else:
            array_cp = np.maximum(k - array_s_nodes, np.zeros(int(n) + 1))
        # check s payoff, according to barriers, if values are different from none
        if h_upper != None:
            array_cp[array_s_nodes >= h_upper] = 0
        if h_lower != None:
            array_cp[array_s_nodes <= h_lower] = 0
        # step backwards recursion through tree
        for i in np.arange(int(n) - 1, -1, -1):
            array_s_nodes = s * d**(np.arange(i, -1, -1)) * \
                u**(np.arange(0, i + 1, 1))
            array_cp[:i + 1] = disc * \
                (q * array_cp[1: i + 2] + (1 - q) * array_cp[0: i + 1])
            array_cp = array_cp[:-1]
            #   evaluating maximum value between fair and intrinsic value
            if opt_style == 'call':
                array_cp = np.maximum(array_cp, array_s_nodes - k)
            else:
                array_cp = np.maximum(array_cp, k - array_s_nodes)
        # returning the no-arbitrage price at node 0
        return array_cp[0]

    def barone_adesi_whaley(self, ):
        '''
        REFERENCES: THE COMPLETE GUIDE TO OPTION PRICING FORMULAS - ESPEN GAARDER HAUG - PG 119
        DOCSTRING: QUADRATIC APPROXIMATION METHOD BY BARONE-ADESI AND WHALEY (1987) TO PRICE 
            CALL AND PUT OPTIONS ON AN UNDERLYING ASSET WITH THE COST-OF-CARRY RATE B - WHEN THE 
            B >= R, THE AMERICAN CALL VALUE IS EQUAL TO THE EUROPEAN CALL VALUE AND CAN THEN BE FOUND 
            BY USING THE GENERALIZED BLACK-SHCOLES-MERTHON (BSM) FORMULA
        INPUTS: PSI_R (VOLATILITY OF REFERENCE VOLATILITY), DICTIONARIES (KEYS - 
            NUMBER OF CONTRACTS WITH VEGA I, T (q_vega), VEGA I, T (vega_t), PSI T (VOLATILITY OF 
            VOLATILITY FOR THE GIVEN SET OF OPTIONS WITH SAME UNDERLYING AND MATURITY) (psi_t), 
            CORRELATION BETWEEN THE VOLATILITY WITH TIME TO MATURITY T AND THE REFERENCE VOLATILITY 
            (corr_t))
        OUTPUTS: FLOAT (VEGA)
        '''



### FORMULAS FOR CALCULATING A STOCK RETURN ###

import numpy as np
import sys
sys.path.append(r'C:\Users\Guilherme\OneDrive\Dev\Python\Packages')
from stpstone.finance.performance_apprraisal.financial_math import FinancialMath


class ValuingStocks:

    def continuous_return(self, stock_d0, stock_d1):
        '''
        DOCSTRING: CALCULATING LN RETURN FOR A STOCK - CONTINUOS RETURN
        INPUTS: STOCK D0 VALUE AND D1 VALUE
        OUTPUTS: FLOAT LN RETURN
        '''
        stock_d0 = float(stock_d0)
        stock_d1 = float(stock_d1)
        return np.log(stock_d1 / stock_d0)

    def discrete_return(self, stock_d0, stock_d1):
        '''
        DOCSTRING: CALCULATING STANDARD RETURN OF A STOCK BETWEEN TWO DATES - DISCRETE RETURN
        INPUTS: STOCK D0 VALUE AND D1 VALUE
        OUTPUTS: FLOAT RETURN
        '''
        stock_d0 = float(stock_d0)
        stock_d1 = float(stock_d1)
        return stock_d1 / stock_d0 - 1

    def calc_returns_from_prices(self, list_prices, type_return='ln_return'):
        '''
        DOCSTRING: LIST OF RETURS FROM A GIVEN LIST OF PRICES
        INPUTS: LIST OF PRICES AND TYPE OF RETURN CALCULATION (LN_RETURN AS STANDARD)
        OUPUTS: LIST OF RETURNS
        '''
        if type_return == 'ln_return':
            return [ValuingStocks().continuous_return(list_prices[i - 1], list_prices[i])
                    for i in range(1, len(list_prices))]
        elif type_return == 'stnd_return':
            return [ValuingStocks().stnd_return(list_prices[i - 1], list_prices[i])
                    for i in range(1, len(list_prices))]
        else:
            raise Exception(
                'Type of return calculation ought be ln_return or stnd_return')

    def pandas_returns_from_spot_prices(self, df_, col_prices, col_dt_date,
                                        col_lag_close='lag_close',
                                        col_first_occurrence_ticker='first_occ_ticker',
                                        col_stock_returns='returns',
                                        type_return='ln_return'):
        '''
        DOCSTRING: PANDAS RETURNS FROM SPOT PRICES OF SECURITIES WITHIN THE DATAFRAME
        INPUTS: DATAFRAME, COL PRICES, COL DATES (DATETIME FORMAT)
        OUTPUTS: DATAFRAME
        '''
        # creating column with first occurrence of a ticker
        df_[col_first_occurrence_ticker] = np.where(df_[col_dt_date] == np.min(df_[col_dt_date]),
                                                    'OK', 'NOK')
        # creating column with lag prices, in order to calculate return
        df_[col_lag_close] = df_[col_prices].shift(periods=-1)
        # calculating returns (continuous or dicrete methods allowed)
        if type_return == 'ln_return':
            df_[col_stock_returns] = \
                df_.apply(lambda row: np.log(row[col_prices] / row[col_lag_close])
                          if row[col_first_occurrence_ticker] == 'NOK' else 0, axis=1)
        else:
            df_[col_stock_returns] = \
                df_.apply(lambda row: row[col_prices] / row[col_lag_close] - 1.0
                          if row[col_first_occurrence_ticker] == 'NOK' else 0, axis=1)
        # returning dataframe
        return df_

    def short_fee_cost(self, fee_short, nper_cd, short_price, quantities, year_cd=360):
        '''
        DOCSTRING: SHORT STRATEGY FEE COST
        INPUTS: FEE, NPER CALENDAR DAYS, SHORT PRICE, QUANTITITES, YEAR CALENDAR DAYS
        OUTPUTS: FLOAT
        '''
        return FinancialMath().compound_interest(fee_short, nper_cd,
                                                 year_cd) * short_price * quantities

    def pricing_strategy(self, long_price, short_price, leverage, operational_costs=0,
                         type_return='ln_return'):
        '''
        DOCSTRING: PNL STOCK STRATEGIES (BUY & HOLD, LONG & SHORT)
        INPUTS: LONGE PRICE, SHORT PRICE, LEVERAGE, OPERATIONAL COSTS
        OUPUTS: DICTIONARY (MTM, PERCENTAGE RETURN, NOTIONAL)
        '''
        if type_return == 'ln_return':
            return {
                'mtm': (float(short_price) - float(long_price)) * float(leverage) - float(
                    operational_costs),
                'pct_retun': ValuingStocks().continuous_return(float(short_price), float(long_price)),
                'notional': float(short_price)
            }
        elif type_return == 'stnd_return':
            return {
                'mtm': (float(short_price) - float(long_price)) * float(leverage) - float(
                    operational_costs),
                'pct_retun': float(long_price) / float(short_price) - 1,
                'notional': float(short_price)
            }
        else:
            raise Exception(
                'Type of return calculation ought be ln_return or stnd_return')


# print(ValuingStocks().calc_returns_from_prices(
#     np.array([2.13, 5.7, 3.1, 4.35, 9.2, 3.7, 4.8])))
# # output
# [0.9843441951191709, -0.6090640633494039, 0.3387737336094921, 0.7490276389544019, -0.9108706644048157, 0.26028309826366636]



### GENERIC FUNCTIONS TO HANDLE POOL CONNECTIONS ###

from typing import Dict, Any


# decorators
def singleton(cls):
    '''
    DOCSTRING: TRACKS WHETER OR NOT AN INSTANCE OF A CLASS IS ALREADY CREATED, KILLING THE OLD ONE 
        IF SO
    INPUTS: CLASS
    OUTPUTS: INSTANCE
    '''
    dict_instances = dict()
    def get_instance(*args, **kwargs):
        if cls not in dict_instances:
            dict_instances[cls] = cls(*args, **kwargs)
        return dict_instances[cls]
    return get_instance


class GenericSQL:

    def format_query(self, query_path:str, dict_params:Dict[str, Any]):
        '''
        DOCSTRING: FORMATS A QUERY WITH F-STRINGS
        INPUTS: QUERY PATH, DICT OF PARAMETERS
        OUTPUTS: FORMATTED QUERY
        '''
        with open(query_path, 'r') as query:
            query_read = query.read()
        return query_read.format(**dict_params)


### HANDLING RISK OF ASSETS TRADED IN EXCHANGE MARKETS ###

import math
import numpy as np
import pandas as pd
import cvxopt as opt
import plotly.graph_objs as go
from itertools import combinations
from datetime import datetime
from typing import List, Tuple, Optional, Dict, Any
from stpstone.quantitative_methods.prob_distributions import NormalDistribution
from stpstone.quantitative_methods.linear_algebra import LinearAlgebra
from stpstone.handling_data.json import JsonFiles
from stpstone.finance.spot.stocks import ValuingStocks
from stpstone.handling_data.lists import HandlingLists
from stpstone.handling_data.numbers import NumHandler
from stpstone.finance.b3.search_by_trading import TradingFilesB3


class MarketRiskManagement:

    def prices_base_normalizer(self, list_returns, base=100):
        '''
        DOCSTRING: RETURNS AS PRICES, PRESERVING DISTANCES THAT WOULD PRESENT THE GIVEN RETURNS
        INPUTS: RETURNS AND BASE
        OUTPUTS: ARRAY
        '''
        # Converts returns into prices
        s = [base]
        for i in range(len(list_returns)):
            s.append(base * (1 + list_returns[i]))
        return np.array(s)

    def beta(self, list_returns, list_market):
        '''
        REFERENCES: http://www.turingfinance.com/computational-investing-with-python-week-one/
        DOCSTRING: BETA OR CORRELATION TO A MARKET BENCHMARK
        INPUTS: LIST OF STOCK RETURNS AND MARKET RETURNS WITH THE SAME SIZE THROUGH THE SAME PERIOD
        OUTPUTS: FLOAT
        '''
        # Create a matrix of [returns, market]
        m = np.matrix([list_returns, list_market])
        # Return the covariance of m divided by the standard deviation of the market returns
        return np.cov(m)[0][1] / np.std(list_market)

    def lower_partial_moment(self, list_returns, threshold, order=2):
        '''
        REFERENCES: http://www.turingfinance.com/computational-investing-with-python-week-one/,
            https://breakingdownfinance.com/finance-topics/performance-measurement/lower-partial-moment/
        DOCSTRING: A MEASURE OF DOWNSIDE RISK COMPUTED AS THE AVERAGE OF THE SQUARED DEVIATIONS
            BELOW A TARGET RETURN. THIS MEASURE OF DOWNSIDE RISK IS MORE GENERAL THAN SEMI-
            VARIANCE WHICH IS COMPUTED AS THE AVERAGE OF THE SQUARED DEVIATIONS BELOW THE MEAN
            RETURN
        INPUTS: RETURNS, THRESHOLD AND ORDER (2 AS DEFAULT)
        OUTPUTS: FLOAT
        '''
        # this method returns a lower partial moment of the returns
        # create an array he same length as returns containing the minimum return threshold
        threshold_array = np.empty(len(list_returns))
        threshold_array.fill(threshold)
        # calculate the difference between the threshold and the returns
        diff = threshold_array - list_returns
        # set the minimum of each to 0
        diff = diff.clip(min=0)
        # return the sum of the different to the power of order
        return np.sum(diff ** order) / len(list_returns)

    def higher_partial_moment(self, list_returns, threshold, order):
        '''
        DOCSTRING: A MEASURE OF DOWNSIDE RISK COMPUTED AS THE AVERAGE OF THE SQUARED DEVIATIONS
            ABOVE A TARGET RETURN. THIS MEASURE OF DOWNSIDE RISK IS MORE GENERAL THAN SEMI-
            VARIANCE WHICH IS COMPUTED AS THE AVERAGE OF THE SQUARED DEVIATIONS BELOW THE MEAN
            RETURN
        INPUTS: RETURNS, THRESHOLD AND ORDER (2 AS DEFAULT)
        OUTPUTS: FLOAT
        '''
        # this method returns a higher partial moment of the returns
        # create an array he same length as returns containing the minimum return threshold
        threshold_array = np.empty(len(list_returns))
        threshold_array.fill(threshold)
        # calculate the difference between the returns and the threshold
        diff = list_returns - threshold_array
        # set the minimum of each to 0
        diff = diff.clip(min=0)
        # return the sum of the different to the power of order
        return np.sum(diff ** order) / len(list_returns)

    def parametric_var(self, std_deviation, t, confidence_level, float_mu=0, h=252, ro=None,
                       port_financial_value=1000):
        '''
        DOCSTRING: VALUE AT RISK, OR THE STATISTICAL RISK MANAGEMENT TECHINIQUE MEASURING THE
            MAXIMUM LOSS THAT AN INVESTMENTT PORTFOLIO IS LIKELY TO FACE WITHIN A SPECIFIC TIME
            FRAME WITH A CERTAIN DEGREE OF CONFIDENCE
        INPUTS: PORTFOLIO FINANCIAL VALUE, STANDARD DEVIATION, T (NPER), COFIDENCE LEVEL (1-ALFA,
            OR SIGNIFICANCE LEVEL), MU (0 AS DEFAULT), H (NOMINAL TIME HORIZON FOR SCALING PURPOSES,
            NUMBER OF WORKING DAYS IN A YEAR, 252 AS DEFAULT) AND RO (AUTOCORRELATION,
            NONE AS DEFAULT)
        OUTPUTS: FINANCIAL POTENCIAL LOSS AND ITS PERCENTAGE OF THE PORTFOLIO (DICTIONARY WITH
            FINANCIAL VAR, PERCENTUAL VAR AND DRIFTED ADJUSTMENT, OR ERROR, THAT WILL BE POSITIVE
            IF THE EXPECTED PORTFOLIO RETURN IS GREATER THAN RISK FREE RATE OF RETURN, AND NEGATIVE
            OTHERWISE)
        '''
        # wheter or not to consider a correlation to correct the transformation among risk
        #   time periods
        if ro:
            # calculating the scaling time horizon
            h_corrected = h + 2 * ro * \
                (1 - ro) ** (-2) * ((h - 1) * (1 - ro) - ro * (1 - ro ** (h - 1)))
        else:
            h_corrected = t / h
        # returning result of parametric var
        return {
            'financial_var': port_financial_value * (std_deviation * np.sqrt(
                h_corrected) * NormalDistribution().inv_norm(
                confidence_level) - (t / h) * float_mu),
            'percentual_var': std_deviation * np.sqrt(
                h_corrected) * NormalDistribution().inv_norm(
                confidence_level) - (t / h) * float_mu,
            'drifted_adjustment_to_var': std_deviation * np.sqrt(
                h_corrected) * NormalDistribution().inv_norm(
                confidence_level) - (std_deviation * np.sqrt(
                    h_corrected) * NormalDistribution().inv_norm(
                    confidence_level) - (t / h) * float_mu)
        }

    def equity_var(self, std_deviation, t, confidence_level, list_betas,
                   list_financial_exposures, float_mu=0, h=252, ro=None):
        '''
        DOCSTRING: EQUITY VAR, OR SENSITIVITY OF ITS RISK FACTORS PORTFOLIO
        INPUTS: STANDARD DEVIATION, T (NPER), COFIDENCE LEVEL (1-ALFA, OR SIGNIFICANCE LEVEL),
            MU (0 AS DEFAULT), H (NOMINAL TIME HORIZON FOR SCALING PURPOSES, NUMBER OF WORKING
            DAYS IN A YEAR, 252 AS DEFAULT) AND RO (AUTOCORRELATION, NONE AS DEFAULT)
        OUTPUTS: DICTIONARY WITH FINANCIAL VAR AND PERCENTUAL VAR
        '''
        # calculating the beta of the portfolio
        beta_portfolio = NumHandler().sumproduct(list_betas, list_financial_exposures) \
            / sum(list_financial_exposures)
        # returning the expected loss of the portfolio regarding a not estressed scenario
        return {
            'financial_var': MarketRiskManagement().parametric_var(
                std_deviation, t, confidence_level, float_mu, h, ro)[
                    'percentual_var'] * beta_portfolio * sum(list_financial_exposures),
            'percentual_var': MarketRiskManagement().parametric_var(
                std_deviation, t, confidence_level, float_mu, h, ro)['percentual_var'] * beta_portfolio
        }

    def normal_linear_interest_rate_var(self, confidence_level, array_pv01, array_yields, t, h,
                                        correl_yields=0, float_mu=0):
        '''
        DOCSTRING: LINEAR RISK CASH FLOW MODEL, OR CASH FLOW MAP, IS THE INTEREST RATE VAR OF
            BONDS, SWAPS AND LOAN PORTFOLIOS THAT CAN BE REPRESENTED AS A SERIES OF CASH FLOWS
        INPUTS: CONFIDENCE LEVEL, ARRAY PV01, ARRAY DERIVATIVE PV01, ARRAY YIELDS, MU (0 AS DEFAULT)
        OUPUTS: FLOAT
        '''
        # covariance matrix of array of yields
        cov_mtx = np.zeros((len(array_yields), len(array_yields)))
        for i in range(len(array_yields)):
            for j in range(len(array_yields)):
                if i == j:
                    cov_mtx[i, j] = array_yields[i] ** (len(array_yields))
                else:
                    cov_mtx[i, j] = correl_yields * \
                        array_yields[i] * array_yields[j]
        cov_mtx = (t / h) * cov_mtx
        # float_mu vector
        float_mu = np.array([float_mu] * len(array_pv01))
        # returning the normal linear interest rate var
        return NormalDistribution().inv_norm(confidence_level) * np.sqrt(
            LinearAlgebra().matrix_multiplication(LinearAlgebra().matrix_multiplication(
                array_pv01, cov_mtx), LinearAlgebra().transpose_matrix(array_pv01))) \
            + LinearAlgebra().matrix_multiplication(float_mu, LinearAlgebra().transpose_matrix(
                array_pv01))

    def expected_tail_loss(self, nth_lowest=10, port_financial_value=1000,
                           list_prices=None, list_returns=None):
        '''
        DOCSTRING: EXPECTED TAIL LOSS FOR A NTH LOWEST AVERAGE LIST OF RETURNS
        INPUTS: NTH LOWEST (10 AS DEFAULT), PORTFOLIO FINANCIAL VALUE (1000 AS VALUE), 
            LIST OF PRICES (NONE AS DEFAULT), LIST OF RETURNS (NONE AS DEFAULT)
        OUTPUTS: DICTIONAY (PERCENTAGE ETL AND FINANCIAL ETL)
        '''
        # checking parameters
        if any(all([x != None for x in [list_prices, list_returns]]),
               all([x == None for x in [list_prices, list_returns]])):
            raise Exception('Please revisit original prices and returns arguments; only one '
                            'of them ought be different from none')
        # defining returns
        if list_prices:
            list_cacl_returns = ValuingStocks().calc_returns_from_prices(list_prices)
        elif list_returns:
            list_cacl_returns = list_returns
        else:
            raise Exception('Please revisit original prices and returns arguments; only one '
                            'of them ought be different from none')
        # nth lowest returns
        return {
            'percentage_etl': np.average(HandlingLists().nth_smallest_numbers(
                list_cacl_returns, nth_lowest)),
            'financial_etl': np.average(HandlingLists().nth_smallest_numbers(
                list_cacl_returns, nth_lowest)) * port_financial_value
        }

    def drawdown(self, tau, list_original_prices=None, list_returns=None, prices_base=100):
        '''
        DOCSTRING: DRAWDOWN, OR HIGHEST DECREASE OF RETURN FOR A GIVEN PORTFOLIO, AMID A MOVING
            TIME RANGE, PRESERVING ITS SIZE
        INPUTS: RETURNS, TAU (TIME PERIOD) AND PRICES BASE (100 AS DEFAULT)
        OUTPUTS: FLOAT
        '''
        # returns the drawdown given time period tau
        if list_returns:
            values = MarketRiskManagement().prices_base_normalizer(list_returns, prices_base)
        if list_original_prices:
            values = list_original_prices
        if all([x != None for x in [list_original_prices, list_returns]]):
            raise Exception('Please revisit original prices and returns arguments; only one '
                            'of them ought be different from none')
        pos = len(values) - 1
        pre = pos - tau
        drawdown = float('+inf')
        # find the maximum drawdown given tau
        while pre >= 0:
            dd_i = (values[pos] / values[pre]) - 1
            if dd_i < drawdown:
                drawdown = dd_i
            pos, pre = pos - 1, pre - 1
        # drawdown should be positive
        return abs(drawdown)

    def max_drawdown(self, list_original_prices=None, list_returns=None, prices_base=100):
        '''
        DOCSTRING: MAXIMUM DRAW DOWN FOR A GIVEN PORTFOLIO
        INPUTS: RETURNS
        OUTPUTS: FLOAT
        '''
        # returns the maximum drawdown for any tau in (0, T) where T is the length of
        #   the return series
        max_drawdown = float('-inf')
        # number of occurrencies in the sample
        if list_original_prices:
            len_observations_data = len(list_original_prices)
        elif list_returns:
            len_observations_data = len(list_returns)
        else:
            raise Exception(
                'List of original prices or returns ought be provided')
        for i in range(len_observations_data):
            drawdown_i = MarketRiskManagement().drawdown(i, list_original_prices,
                                                         list_returns, prices_base)
            if drawdown_i > max_drawdown:
                max_drawdown = drawdown_i
        # max drawdown should be positive
        return abs(max_drawdown)

    def ewma(self, list_daily_returns, int_wdy=252, accuracy=0.01):
        '''
        REFERENCES: https://www.investopedia.com/articles/07/ewma.asp
        DOCSTRING: EXPONENTIALLY WEIGHTED MOVING AVERAGE
        INPUTS: DAILY RETURNS, LAMBDA SMOOTHING PARAMETER (BY DEFAULT 94%, ACCORDING TO
            RISKMETRICS)
        OUTPUTS:
        '''
        # number of observations precision-wise
        list_accuracy_0_01 = [44, 49, 55, 63, 74, 90, 113, 151, 228, 458]
        list_accuracy_0_001 = [66, 73, 83, 95, 112, 135, 169, 227, 342]
        list_accuracy_0_0001 = [87, 98, 110, 127, 149, 180, 226, 302, 456, 916]
        list_lambdas = [x / 100 for x in range(90, 100)]
        # lambda due to precision and number of observations
        curr_accuracy_list = locals()[
            'list_accuracy_' + str(accuracy).replace('.', '_')]
        # finding lambda according to number of observations and accuracy
        if accuracy in [0.01, 0.001, 0.0001]:
            if len(list_daily_returns) > curr_accuracy_list[-1]:
                lamba_smoothing_parameter = list_lambdas[-1]
            else:
                lamba_smoothing_parameter = list_lambdas[curr_accuracy_list.index(
                    HandlingLists().closest_bound(curr_accuracy_list, len(list_daily_returns)))]
        else:
            return 'Poor defined variable accuracy, please make sure its within ' \
                + '[0.01, 0.001, 0.0001] possible values'
        # periodic returns
        u_ln_returns = [ValuingStocks().continuous_return(
            list_daily_returns[p - 1],
            list_daily_returns[p]) for p in range(1, len(list_daily_returns))]
        # array_weights
        array_weights = [(1 - lamba_smoothing_parameter) * lamba_smoothing_parameter ** (d - 1) for
                         d in range(1, len(list_daily_returns))]
        array_weights.reverse()
        # ewma daily contributions
        if len(u_ln_returns) == len(array_weights):
            ewma_daily_contributions_list = [array_weights[i] * u_ln_returns[i] ** 2 for i in
                                             range(len(u_ln_returns))]
        else:
            return 'Please check wheter daily returns and array_weights list are the same size and ' \
                'try again.'
        # variance and standard deviation ewma-wise
        dict_message = {
            'variance_ewma_daily': sum(ewma_daily_contributions_list),
            'std_ewma_daily': math.sqrt(sum(ewma_daily_contributions_list)),
            'variance_ewma_yearly': (math.sqrt(sum(
                ewma_daily_contributions_list)) * math.sqrt(int_wdy)) ** 2,
            'std_ewma_yearly': math.sqrt(sum(
                ewma_daily_contributions_list)) * math.sqrt(int_wdy)
        }
        return JsonFiles().send_json(dict_message)

    def systematic_specific_risk(self, variance_portfolio, std_market, array_stocks_weights,
                                 array_stocks_beta):
        '''
        DOCSTRING: SYSTEMATIC AND SPECIFIC RISK FOR A GIVEN PORTFOLIO
        INPUTS: STD PORTFOLIO, ARRAY STOCKS WEIGHTS AND ARRAY STOCKS BETA
        OUTPUTS: DICTIONARY WITH STD PORTFOLIO, SYSTEMATIC RISK AND SPECIFIC RISK
        '''
        beta_portfolio = \
            LinearAlgebra().matrix_multiplication(np.array(array_stocks_weights),
                                                  LinearAlgebra().transpose_matrix(
                                                      np.array(array_stocks_beta)))
        systematic_risk = math.pow(beta_portfolio, 2) * \
            math.pow(std_market, 2)
        specific_risk = variance_portfolio - systematic_risk
        return {
            'variance_portfolio': variance_portfolio,
            'systematic_risk': systematic_risk,
            'specifici_risk': specific_risk
        }


class MarkowitzEff:
    '''
    REFERENCES: https://www.linkedin.com/pulse/python-aplicado-markowitz-e-teoria-nem-tão-moderna-de-paulo-rodrigues/?originalSubdomain=pt
    DOCSTRING: MARKOWITZ RISK-RETURN PLOT OF RANDOM PORTFOLIOS, AIMING TO FIND THE BEST ALLOCATION 
        WITH THE ASSETS PROVIDED
    INPUTS: -
    OUTPUTS: -
    '''

    def __init__(
            self, df_mktdata:pd.DataFrame, int_n_portfolios:int, float_prtf_notional:float, 
            float_rf:float, col_ticker:str='ticker', col_close:str='close', col_dt:datetime='dt_date', 
            col_returns:str='daily_return', col_last_close:str='last_close', col_min_w:str='min_w', 
            col_max_date:str='max_date', bl_constraints:bool=True, bl_opt_possb_comb:bool=False, 
            bl_progress_printing_opt:bool=False, nth_try:str=100, n_attempts_opt_prf:int=10000, 
            int_wdy:int=252, int_round_close:int=2, path_fig:Optional[str]=None, 
            bl_debug_mode:bool=True, bl_show_plot:bool=True, bl_non_zero_w_eff:bool=False, 
            title_text:str='Markowitz Risk x Return Portfolios',
            yaxis_title:str='Return (%)', xaxis_title:str='Risk (%)'
        ) -> None:
        # attributes
        self.float_prtf_notional = float_prtf_notional
        self.df_mktdata = df_mktdata
        self.path_fig = path_fig
        self.int_n_portfolios = int_n_portfolios
        self.n_attempts_opt_prf = n_attempts_opt_prf
        self.col_ticker = col_ticker
        self.col_close = col_close
        self.col_dt = col_dt
        self.col_returns = col_returns
        self.col_min_w = col_min_w
        self.col_last_close = col_last_close
        self.col_max_date = col_max_date
        self.float_rf = float_rf
        self.bl_constraints = bl_constraints
        self.bl_opt_possb_comb = bl_opt_possb_comb
        self.bl_progress_printing_opt = bl_progress_printing_opt
        self.nth_try = nth_try
        self.int_wdy = int_wdy
        self.int_round_close = int_round_close
        self.bl_debug_mode = bl_debug_mode
        self.bl_show_plot = bl_show_plot
        self.bl_non_zero_w_eff = bl_non_zero_w_eff
        self.title_text = title_text
        self.yaxis_title = yaxis_title
        self.xaxis_title = xaxis_title
        # securities
        self.list_securities = df_mktdata[self.col_ticker].unique()
        # last date per asset
        self.df_mktdata[self.col_max_date] = self.df_mktdata.groupby([
            self.col_ticker])[self.col_dt].transform('max')
        # minimum w per asset
        if self.col_min_w not in self.df_mktdata.columns:
            self.df_mktdata[self.col_min_w] = self.df_mktdata.groupby([
                self.col_ticker, 
                self.col_max_date
            ])[col_close].transform('last') / self.float_prtf_notional
        # last closing prices
        self.df_mktdata[self.col_last_close] = self.df_mktdata.groupby([
            self.col_ticker, 
            self.col_max_date
        ])[col_close].transform('last')
        self.array_close = np.array([
            self.df_mktdata[self.df_mktdata[self.col_ticker] == ticker.replace('.SA', '')][
                self.col_last_close].unique()[0]
            for ticker in self.list_securities
        ])
        # random portfolios
        self.array_mus, self.array_sigmas, self.array_sharpes, self.array_weights, \
            self.array_returns, self.list_uuids = \
                self.random_portfolios(
                    self.df_mktdata, int_n_portfolios, self.col_ticker, self.col_close, 
                    self.col_dt, self.col_returns, self.float_prtf_notional, self.col_min_w, 
                    self.float_rf, self.bl_constraints, self.bl_opt_possb_comb, self.nth_try, 
                    self.int_wdy
                )
        # optimal portfolios
        self.array_eff_weights, self.array_eff_returns, self.array_eff_risks = \
            self.optimal_portfolios(
                self.array_returns, self.n_attempts_opt_prf, self.bl_progress_printing_opt, 
                self.int_wdy
            )
        # efficient frontier
        self.df_eff, self.df_porf = self.eff_frontier(
            self.array_eff_risks, self.array_eff_returns, self.array_weights, self.array_mus, 
            self.array_sigmas, self.float_rf
        )

    def sharpe_ratio(self, float_mu:float, float_sigma:float, float_rf:float) -> float:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return (float(float_mu) - float(float_rf)) / float(float_sigma)

    def sigma_portfolio(self, array_weights:np.array, array_returns:np.array) -> float:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # covariance between stocks
        array_cov = np.cov(array_returns)
        # returning portfolio standard deviation
        return np.sqrt(np.dot(array_weights.T, np.dot(array_cov, array_weights)))

    def returns_min_w_uids(self, df_assets:pd.DataFrame, col_dt:str, col_id:str, 
                           col_returns:str, col_min_w:str) \
                            -> Tuple[np.ndarray, np.ndarray, List[str]]:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # filter where returns are not nulls
        df_assets = df_assets[~df_assets[col_returns].isnull()]
        # returns per uids
        array_returns = df_assets.pivot_table(
            index=col_dt,
            columns=col_id,
            values=col_returns
        ).to_numpy()
        array_returns = array_returns.T
        array_returns = np.nan_to_num(array_returns, nan=0.0)
        # minimum weights per uids
        array_min_w = np.array(
            df_assets.groupby(
                col_id
            )[col_min_w].unique(),
            dtype=float
        )
        array_min_w = array_min_w.T
        # list of uids
        list_uids = HandlingLists().remove_duplicates(df_assets[col_id].to_list())
        # return arrays of interet
        return array_returns, array_min_w, list_uids

    def random_weights(self, int_n_assets:int, bl_constraints:bool=False, bl_opt_possb_comb:bool=False, 
                       array_min_w:np.array=None, nth_try:int=100, int_idx_val:int=2, 
                       bl_valid_weights:bool=False, i_attempts:int=0, 
                       float_atol_sum:float=1e-4, float_atol_w:float=10000.0) -> np.array:
        '''
        DOCSTRING: RANDOM WEIGHTS - WITH OR WITHOUT CONSTRAINTS
        INPUTS:
            - INT_N_ASSETS: THE NUMBER OF ASSETS IN THE PORTFOLIO
            - BL_CONSTRAINTS: BOOLEAN FLAG TO APPLY CONSTRAINTS OR NOT
            - MIN_INVEST_PER_ASSET: A LIST OF MINIMUM WEIGHTS/INVESTMENTS FOR EACH ASSET
        OUTPUTS:
            - A LIST OF WEIGHTS FOR THE ASSETS THAT SATISFY THE GIVEN CONSTRAINTS, 
                WHERE SUM OF WEIGHTS = 1
        '''
        # adjusting number of assets within the portfolio
        int_idx_val = min(len(array_min_w), int_idx_val)
        # check wheter the constraints are enabled
        if bl_constraints == True:
            #   sanity check for constraints
            if array_min_w is None:
                raise ValueError('MIN_INVEST_PER_ASSET MUST BE PROVIDED AS A LIST WHEN '
                                 + 'CONSTRAINTS ARE ENABLED.')
            if any(isinstance(x, str) for x in array_min_w):
                raise ValueError('MIN_INVEST_PER_ASSET MUST BE A LIST OF NUMBERS.')
            if len(array_min_w) != int_n_assets:
                raise ValueError('THE LENGTH OF MIN_INVEST_PER_ASSET MUST MATCH THE '
                                 + 'NUMBER OF ASSETS.')
            if any(x < 0 for x in array_min_w):
                raise ValueError('MIN_INVEST_PER_ASSET MUST BE POSITIVE.')
            if any(x > 1 for x in array_min_w):
                raise ValueError('MIN_INVEST_PER_ASSET MUST BE BELOW 1.0')
            if any(x == 0 for x in array_min_w):
                raise ValueError('EVERY MIN_INVEST_PER_ASSET MUST BE GREATER THAN 0.')
            #   initializing variables
            bl_valid_weights = False
            list_combs = [
                comb 
                for r in range(2, int_idx_val + 1) 
                for comb in combinations(array_min_w, r)
            ]
            #   recursive call to get valid weights
            while not bl_valid_weights:
                #   increment the try counter
                i_attempts += 1
                #   reseting variables
                array_w = np.zeros(int_n_assets)
                #   check if it's the nth try or all the combinations are greater than one
                if \
                    (i_attempts >= nth_try)\
                    or (all([sum(comb) >= 1.0 for comb in list_combs])):
                    #   return a weight array with one asset having weight 1.0 and others 0.0
                    array_w = np.zeros(int_n_assets)
                    int_idx = np.random.randint(0, int_n_assets)
                    array_w[int_idx] = 1.0
                    return array_w
                #   if multiplier is enabled, build a list of possible indexes combinations in 
                #       order to sum 1.0 or less
                if bl_opt_possb_comb == True:
                    #   combinations where sum is less than 1.0 - flatten list
                    list_i = HandlingLists().remove_duplicates([
                        idx
                        for comb in list_combs
                        for x in comb
                        for idx in np.where(array_min_w == x)[0]
                        if sum(comb) <= 1.0
                    ])
                else:
                    list_i = list(range(int_n_assets))
                np.random.shuffle(list_i)
                #   looping through the indexes
                for i in list_i:
                    #   randomly building a float weight
                    float_upper_tol = max(
                        float_atol_w * (1.0 - sum(array_w)), 
                        1.0
                    )
                    #   building the float weight with any given value above the minimum or a 
                    #       multiple of the minimum
                    if bl_opt_possb_comb == True:
                        int_max_mult = max(
                            int((1.0 - sum(array_w)) // array_min_w[i]), 
                            1
                        )
                        int_rand_mult = np.random.randint(0, int_max_mult + 1)
                        float_weight = float(int_rand_mult * array_min_w[i])
                    else:
                        float_upper_tol = max(
                            float_atol_w * (1.0 - sum(array_w)), 
                            1.0
                        )
                        random_integer = np.random.randint(
                            0,
                            float_upper_tol
                        )
                        float_weight = float(random_integer) / float_upper_tol
                    #   check if the weight is greater than the minimum
                    if float_weight < array_min_w[i]:
                        array_w[i] = 0
                    else:
                        array_w[i] = float_weight
                    #   check if the sum of weights is equal to 1.0 or greater
                    if sum(array_w) >= 1.0: break
                #   normalize only if the total weight is non-zero, if multiplier is unabled
                if (bl_opt_possb_comb == False) or (np.count_nonzero(array_w) == 1):
                    total_weight = np.sum(array_w)
                    if total_weight > 0:
                        array_w /= total_weight
                #   sanity checks for weights:
                #       1 - all weights must be non-negative
                #       2 - sum must be equal to 1
                #       3 - the minimum must be respected, or zero for a given asset
                #       4 - some weight must be positive
                bl_valid_weights = (
                    np.all(array_w >= 0) 
                    and np.isclose(np.sum(array_w), 1, atol=float_atol_sum) 
                    and all([
                        (array_w[i] >= array_min_w[i]) \
                            or (array_w[i] == 0)
                        for i in range(int_n_assets)
                    ]) 
                    and np.any(array_w > 0)
                    and np.all(array_w != 1)
                )
            return array_w
        else:
            # if no constraints are applied, return standard random weights
            k = np.random.rand(int_n_assets)
            return k / sum(k)

    def random_portfolio(self, array_returns:np.ndarray, float_rf:float, bl_constraints:bool=False, 
                        bl_opt_possb_comb:bool=False, array_min_w:Optional[np.ndarray]=None, 
                        nth_try:int=100, int_wdy:int=252) \
                            -> Tuple[np.ndarray, np.ndarray, np.ndarray, str]:
        '''
        DOCSTRING: RETURNS THE MEAN AND STANDARD DEVIATION OF RETURNS FROM A RANDOM PORTFOLIO
        INPUTS: MATRIX ASSETS RETURNS, ARRAY EXPECTED RETURNS, FLOAT RISK FREE
        OUTPUTS: TUP OF FLOATS
        '''
        # adjusting variables' types
        array_r = np.asmatrix(array_returns)
        float_rf = float(float_rf)
        # random wieghts for the current portfolio
        array_weights = self.random_weights(array_r.shape[0], bl_constraints, bl_opt_possb_comb, 
                                            array_min_w, nth_try)
        # mean returns for assets
        array_returns = np.asmatrix(np.mean(array_r, axis=1))
        # portfolio standard deviation
        array_sigmas = self.sigma_portfolio(array_weights, array_r) * np.sqrt(int_wdy)
        # portfolio expected return
        array_mus = float(array_weights * array_returns) * int_wdy
        # sharpes ratio
        array_sharpes = self.sharpe_ratio(array_mus, array_sigmas, float_rf)
        # changing type of array weights to transform into one value
        array_weights = ' '.join([str(x) for x in array_weights])
        # returning portfolio infos
        return array_mus, array_sigmas, array_sharpes, array_weights

    def random_portfolios(self, df_assets:pd.DataFrame, int_n_portfolios:int, col_id:str, 
                          col_close:str, col_dt:datetime, col_returns:str, 
                          float_prtf_notional:float, col_min_w:str='min_w', float_rf:float=0.0, 
                          bl_constraints:bool=False, bl_opt_possb_comb:bool=False, 
                          nth_try:int=100, int_wdy:int=252) \
                            -> Tuple[np.ndarray, np.ndarray, np.ndarray, 
                                     np.ndarray, np.ndarray, List[str]]:
        '''
        DOCSTRING: RETURNS THE MEAN AND STANDARD DEVIATION OF RETURNS FROM A RANDOM PORTFOLIO
        INPUTS: MATRIX ASSETS RETURNS, ARRAY EXPECTED RETURNS, FLOAT RISK FREE
        OUTPUTS: TUP OF FLOATS
        '''
        # arrays of retunrs and minimum weights per asset
        array_returns, array_min_w, list_uuids = \
            self.returns_min_w_uids(
                df_assets,
                col_dt,
                col_id,
                col_returns,
                col_min_w,
            )
        # generating random portfolios
        array_mus, array_sigmas, array_sharpes, array_weights = \
            np.column_stack([
                self.random_portfolio(
                    array_returns,
                    float_rf,
                    bl_constraints,
                    bl_opt_possb_comb,
                    array_min_w,
                    nth_try,
                    int_wdy
                )
                for _ in range(int_n_portfolios)
            ])
        # altering data types
        array_mus = array_mus.astype(float)
        array_sigmas = array_sigmas.astype(float)
        array_sharpes = array_sharpes.astype(float)
        return array_mus, array_sigmas, array_sharpes, array_weights, array_returns, list_uuids

    def optimal_portfolios(self, array_returns:np.ndarray, n_attempts:int=1000,
                           bl_progress_printing_opt:bool=False, int_wdy:int=252) \
                             -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        '''
        DOCSTRING: WEIGHTS RETURNS AND SIGMA FOR EFFICIENT FRONTIER
        INPUTS: MATRIX OF ASSETS' RETURNS
        OUTPUTS: TUP OF ARRAYS
        '''
        # turn on/off progress printing
        opt.solvers.options['show_progress'] = bl_progress_printing_opt
        # configuring data types
        array_returns = np.asmatrix(array_returns)
        # definig the number of portfolios to be created
        n = array_returns.shape[0]
        # calculating first attempt for float_mu in each portfolio
        mus = [10.0 ** (5.0 * float(t / n_attempts) - 1.0)
               for t in range(n_attempts)]
        # convert to cvxopt matrices
        S = opt.matrix(np.cov(array_returns))
        pbar = opt.matrix(np.mean(array_returns, axis=1))
        # create constraint matrices
        #   negative n x n identity matrix
        G = -opt.matrix(np.eye(n))
        h = opt.matrix(0.0, (n, 1))
        A = opt.matrix(1.0, (1, n))
        b = opt.matrix(1.0)
        # calculate efficient frontier weights using quadratic programming
        list_portfolios = [opt.solvers.qp(float_mu * S, -pbar, G, h, A, b)['x']
                           for float_mu in mus]
        # calculating risk and return for efficient frontier
        array_returns = [opt.blas.dot(
            pbar, x) * int_wdy for x in list_portfolios]
        array_sigmas = [np.sqrt(opt.blas.dot(
            x, S * x)) * np.sqrt(int_wdy) for x in list_portfolios]
        # calculate the second degree polynomial of the frontier curve
        m1 = np.polyfit(array_returns, array_sigmas, 2)
        x1 = np.sqrt(m1[2] / m1[0])
        # calculate the optimal portfolio
        wt = opt.solvers.qp(opt.matrix(x1 * S), -pbar, G, h, A, b)['x']
        # returning weights, returns, and sigma from efficient frontier
        return np.asarray(wt), array_returns, array_sigmas

    def eff_frontier(self, array_eff_risks:np.array, array_eff_returns:np.array, 
                     array_weights:np.array, array_mus:np.array, array_sigmas:np.array, 
                     float_rf:float, col_sigma:str='sigma', col_mu:str='float_mu', 
                     col_w:str='weights', col_sharpe:str='sharpe', 
                     atol:float=1e-2, int_pace_atol:int=5) -> Tuple[pd.DataFrame, pd.DataFrame]:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        array_eff_weights = list()
        # convert string-based array_weights to a 2D array by splitting the values
        array_weights_2d = np.array([list(map(float, row.split())) for row in array_weights])
        # iterate over the efficient returns and risks
        for _, eff_risk in zip(array_eff_returns, array_eff_risks):
            while True:
                try:
                    #   find the indices in sigmas that correspond to the current risk using 
                    #       np.isclose
                    list_idx_sigma = np.where(np.isclose(array_sigmas, eff_risk, atol=atol))
                    #   get the highest return for the given datasets
                    idx_mu = np.argmax(array_mus[list_idx_sigma])
                    # print(eff_risk, list_idx_sigma, array_mus[list_idx_sigma], idx_mu)
                    # raise Exception('BREAK')
                    #   get the index from mus and append weights
                    array_eff_weights.append(array_weights_2d[idx_mu])
                    #   in case of no error break the loop
                    break
                except ValueError:
                    atol *= int_pace_atol
        # convert to numpy array for final output if needed
        array_eff_weights = np.array(array_eff_weights)
        # create a dataframe
        columns = [f'weight_{i}' for i in range(array_eff_weights.shape[1])]
        df_eff = pd.DataFrame(array_eff_weights, columns=columns)
        df_eff[col_mu] = array_eff_returns
        df_eff[col_sigma] = array_eff_risks
        # calculate sharpe as the difference between array_eff_returns and float_rf 
        #   divided by array_eff_risks
        df_eff[col_sharpe] = (df_eff[col_mu] - float_rf) / df_eff[col_sigma]
        # create a pandas dataframe with returns, weights and mus from the original porfolios
        df_porf = pd.DataFrame({col_mu: array_mus, col_sigma: array_sigmas, col_w: array_weights})
        # output the results
        return df_eff, df_porf
    
    @property
    def plot_risk_return_portfolio(self) -> None:
        '''
        REFERENCES: https://plotly.com/python/reference/layout/, 
            https://plotly.com/python-api-reference/generated/plotly.graph_objects.Scatter.html, 
            https://plotly.com/python/builtin-colorscales/
        DOCSTRING: PLOT MARKOWITZ'S EFFICIENT FRONTIER FOR PORTFOLIO MANAGEMENT
        INPUTS: ARRAY WEIGHTS, ARRAY MUS (MEAN RETURNS FOR EACH GIVEN PORTFOLIO, BASED ON EXPCTED 
            RETURNS FOR EACH SECURITY, GIVEN ITS WEIGHT ON THE SYNTHETIC PORTFOLIO), ARRAY OF SHARPES, 
            ARRAY OF EFFECTIVE RISKS, ARRAY OF EFFECTIVE RETURN FOR ALL SECURITIES IN A PORTFOLIO, 
            TITLE, YAXIS NAME AND XAXIS NAME
        OUTPUTS: PLOT
        '''
        # maximum sharpe portfolio
        idx_max_sharpe = self.array_sharpes.argmax()
        max_sharpe_sigma = self.array_sigmas[idx_max_sharpe]
        max_sharpe_mu = self.array_mus[idx_max_sharpe]
        # minimum sigma portfolio
        idx_min_sigma = self.array_sigmas.argmin()
        min_sigma_mu = self.array_mus[idx_min_sigma]
        min_sigma_sigma = self.array_sigmas[idx_min_sigma]
        # maximum sharpe portfolio
        if self.bl_debug_mode == True:
            print('### MAXIMUM SHARPE PORTFOLIO ###')
            print('SHARPES ARGMAX: {}'.format(self.array_sharpes.argmax()))
            print('WEIGHTS: {}'.format(self.array_weights[self.array_sharpes.argmax()]))
            print('RISK: {}'.format(self.array_sigmas[self.array_sharpes.argmax()]))
            print('RETURN: {}'.format(self.array_mus[self.array_sharpes.argmax()]))
            print('SHARPE: {}'.format(self.array_sharpes[self.array_sharpes.argmax()]))
        # prepare customdata for scatter plot
        customdata_portfolios = np.array([
            [weights, ', '.join(self.list_securities)] for weights in self.array_weights
        ], dtype=object)
        # Ppepare the subtitle with the list of securities
        subtitle_text = 'List of securities: ' + ', '.join(self.list_securities)
        # ploting data
        data = [
            go.Scatter(
                x=self.array_sigmas.flatten(),
                y=self.array_mus.flatten(),
                mode='markers',
                marker=dict(
                    color=self.array_sharpes.flatten(),
                    colorscale='Viridis',
                    showscale=True,
                    cmax=self.array_sharpes.flatten().max(),
                    cmin=0,
                    colorbar=dict(
                        title='Sharpe Ratios'
                    )
                ),
                #   define the hovertemplate to include weights
                hovertemplate=(
                    'Risk: %{x:.2f}<br>' +
                    'Return: %{y:.2f}<br>' +
                    'Sharpe: %{marker.color:.2f}<br>' +
                    'Weight: %{customdata[0]}<extra></extra>'
                ),
                #   weights data for hovertemplate
                customdata=customdata_portfolios,
                name='Portfolios'
            ),
            go.Scatter(
                x=self.array_eff_risks,
                y=self.array_eff_returns,
                mode='lines+markers', 
                line=dict(color='red', width=2),
                name='Efficient Frontier',
                hovertemplate=(
                    'Risk: %{x:.2f}<br>' +
                    'Return: %{y:.2f}<br>' +
                    'Weight: %{customdata}<extra></extra>'
                ),
                customdata=self.array_eff_weights
            ),
            # add a green star for the minimum sigma portfolio
            go.Scatter(
                x=[min_sigma_sigma],
                y=[min_sigma_mu],
                mode='markers',
                marker=dict(size=30, color='green', symbol='star'),
                name='Min Risk Portfolio',
                hovertemplate='Risk: %{x:.2f}<br>Return: %{y:.2f}<extra></extra>'
            ),
            # add a red star for the maximum sharpe portfolio
            go.Scatter(
                x=[max_sharpe_sigma],
                y=[max_sharpe_mu],
                mode='markers',
                marker=dict(size=30, color='blue', symbol='star'),
                name='Max Sharpe Portfolio',
                hovertemplate='Risk: %{x:.2f}<br>Return: %{y:.2f}<extra></extra>'
            )
        ]
        # configuring title data
        dict_title = {
            'text': self.title_text,
            'xanchor': 'center',
            'yanchor': 'top',
            'y': 0.95,
            'x': 0.5
        }
        # legend
        dict_leg = {
            'orientation': 'h', 
            'yanchor': 'bottom',
            'y': -0.2,
            'xanchor': 'center', 
            'x': 0.5
        }
        # launching figure with plotly
        fig = go.Figure(data=data)
        # update layout
        fig.update_layout(
            title=dict_title,
            annotations=[
                dict(
                    text=subtitle_text,
                    x=0.53,
                    y=1.08,
                    xref='paper',
                    yref='paper',
                    showarrow=False,
                    font=dict(
                        size=12,
                        color='gray'
                    ),
                    align='center'
                )
            ],
            xaxis_rangeslider_visible=False, width=1280, height=720,
            xaxis_showgrid=True, xaxis_gridwidth=1, xaxis_gridcolor='#E8E8E8',
            yaxis_showgrid=True, yaxis_gridwidth=1, yaxis_gridcolor='#E8E8E8',
            yaxis_title=self.yaxis_title, xaxis_title=self.xaxis_title,
            legend=dict_leg,
            plot_bgcolor='rgba(0,0,0,0)',
        )
        # save plot, if is user's interest
        if self.path_fig is not None:
            fig_extension = self.path_fig.split('.')[-1]
            fig.write_image(
                self.path_fig, 
                format=fig_extension, 
                scale=2, 
                width=1280, 
                height=720
            )
        # display plot
        if self.bl_show_plot == True:
            fig.show()

    @property
    def max_sharpe(self) -> dict:
        '''
        DOCSTRING: MAXIMUM SHARPE RATIO PORTFOLIO
        INPUTS:
            - array_sharpes: np.ndarray -> ARRAY CONTAINING SHARPE RATIOS FOR THE PORTFOLIOS
            - array_weights: np.ndarray -> ARRAY CONTAINING ASSET WEIGHTS IN THE PORTFOLIOS
            - array_sigmas: np.ndarray -> ARRAY CONTAINING RISKS (STANDARD DEVIATIONS) OF THE PORTFOLIOS
            - array_mus: np.ndarray -> ARRAY CONTAINING EXPECTED RETURNS OF THE PORTFOLIOS
            - ensure_nonzero_weights: bool -> ENSURE THAT ALL WEIGHTS ARE NON-ZERO
        OUTPUTS:
            - dict -> DICTIONARY CONTAINING INFORMATION ABOUT THE MAXIMUM SHARPE RATIO PORTFOLIO
        '''
        # ensuring that all weights are non-zero, if is user's interest
        if self.bl_non_zero_w_eff:
            array_valid_indices = np.where((self.array_weights != 0).all(axis=1))[0]
            if len(array_valid_indices) == 0:
                raise ValueError('No available portfolios with non-zero weights')
            int_argmax_sharpe = array_valid_indices[self.array_sharpes[array_valid_indices].argmax()]
        else:
            int_argmax_sharpe = self.array_sharpes.argmax()
        # maximum sharpe ratio portfolio
        array_eff_w = self.array_weights[self.array_sharpes.argmax()]
        array_eff_mu = self.array_mus[self.array_sharpes.argmax()]
        array_eff_sharpe = self.array_sharpes[self.array_sharpes.argmax()]
        # efficient quantities
        array_eff_quantities = [
            round(float(w) * self.float_prtf_notional / self.array_close[i]) 
            for i, w in enumerate(array_eff_w.split())
        ]
        # calculating notional (ensure array_eff_quantities is properly calculated as float)
        self.array_close = np.round(self.array_close, self.int_round_close)
        array_notional = np.array(self.array_close) * np.array(array_eff_quantities)
        return {
            'tickers': self.list_securities,
            'argmax_sharpe': int_argmax_sharpe,
            'eff_weights': array_eff_w,
            'eff_mu': array_eff_mu,
            'eff_sharpe': array_eff_sharpe,
            'eff_quantities': array_eff_quantities,
            'close': self.array_close,
            'notional': array_notional,
            'notional_total': array_notional.sum()
        }

    @property
    def min_sigma(self) -> dict:
        '''
        DOCSTRING: MINIMUM RISK PORTFOLIO
        INPUTS:
            - array_sharpes: np.ndarray -> ARRAY CONTAINING SHARPE RATIOS FOR THE PORTFOLIOS
            - array_weights: np.ndarray -> ARRAY CONTAINING ASSET WEIGHTS IN THE PORTFOLIOS
            - array_sigmas: np.ndarray -> ARRAY CONTAINING RISKS (STANDARD DEVIATIONS) OF THE PORTFOLIOS
            - array_mus: np.ndarray -> ARRAY CONTAINING EXPECTED RETURNS OF THE PORTFOLIOS
            - bl_non_zero_w_eff: bool -> ENSURE THAT ALL WEIGHTS ARE NON-ZERO
        OUTPUTS:
            - dict -> DICTIONARY CONTAINING INFORMATION ABOUT THE MINIMUM RISK PORTFOLIO
        '''
        
        # ensuring that all weights are non-zero, if is user's interest
        if self.bl_non_zero_w_eff:
            array_valid_indices = np.where((self.array_weights != 0).all(axis=1))[0]
            if len(array_valid_indices) == 0:
                raise ValueError('No available portfolios with non-zero weights')
            int_argmin_risk = array_valid_indices[self.array_sigmas[array_valid_indices].argmin()]
        else:
            int_argmin_risk = self.array_sigmas.argmin()
        # minimum risk portfolio
        array_eff_w = self.array_weights[int_argmin_risk]
        array_eff_risk = self.array_sigmas[int_argmin_risk]
        array_eff_mu = self.array_mus[int_argmin_risk]
        # efficient quantities
        array_eff_quantities = [
            round(float(w) * self.float_prtf_notional / self.array_close[i]) 
            for i, w in enumerate(array_eff_w.split())
        ]
        # calculating notional (ensure array_eff_quantities is properly calculated as float)
        self.array_close = np.round(self.array_close, self.int_round_close)
        array_notional = np.array(self.array_close) * np.array(array_eff_quantities)
        return {
            'tickers': self.list_securities,
            'argmin_risk': int_argmin_risk,
            'eff_weights': array_eff_w,
            'eff_risk': array_eff_risk,
            'eff_mu': array_eff_mu,
            'eff_quantities': array_eff_quantities,
            'close': self.array_close,
            'notional': array_notional,
            'notional_total': array_notional.sum()
        }

    def lot_shares_ticker_corr(self, dict_allocation:Dict[str, Any]) \
        -> Dict[str, Any]:
        '''
        DOCSTRING: LOT OF SHARES TICKER CORRECTION
        INPUTS: DICT ALLOCATION
        OUTPUTS: DICT
        '''
        list_ser = list()
        df_trad_sec = TradingFilesB3().tradable_securities
        for ticker in dict_allocation['tickers']:
            float_qty_rnd = df_trad_sec[df_trad_sec['Symbol'] == ticker.replace('.SA', '')][
                'MinOrderQty'].values[0]
            float_qty_quotient = float(dict_allocation['eff_quantities'][
                self.list_securities.index(ticker)]) // float_qty_rnd
            float_qty_remainder = float(dict_allocation['eff_quantities'][
                self.list_securities.index(ticker)]) % float_qty_rnd
            list_ser.append({'ticker': ticker.replace('.SA', ''), 'qty': float_qty_quotient})
            list_ser.append({'ticker': ticker.replace('.SA', '') + 'F', 'qty': float_qty_remainder})
        df_ = pd.DataFrame(list_ser)
        df_ = df_[df_['qty'] != 0]
        df_['close'] = list(dict_allocation['close'])
        df_['notional'] = df_['qty'] * df_['close']
        return df_